from datetime import datetime, date, timedelta


def coerce_timestamp(value):
    return datetime.fromtimestamp(value)


def coerce_interval(value):
    if isinstance(value, (int, float)):
        return timedelta(seconds=value)
    else:
        assert 0


def coerce_date(value):
    return date.fromtimestamp(value)


TYPE_MAPPING = {
    'timestamp': coerce_timestamp,
    'date': coerce_date,
    'interval': coerce_interval
}


def coerce(type, value):
    if value is None:
        return None
    if type not in TYPE_MAPPING:
        return value
    return TYPE_MAPPING[type](value)



import os
from contextlib import contextmanager

from polecat.db.connection import cursor, manager
from polecat.db.utils import unparse_url
from polecat.utils import random_ident


@contextmanager
def environ(**kwargs):
    orig = dict()
    todel = []
    try:
        for k, newval in kwargs.items():
            if k in os.environ:
                orig[k] = os.environ[k]
            else:
                todel.append(k)
            os.environ[k] = newval
        yield
    finally:
        for k, oldval in orig.items():
            os.environ[k] = oldval
        for k in todel:
            del os.environ[k]


@contextmanager
def create_database():
    dbinfo = manager.parse_url()
    test_dbname = random_ident()
    with cursor() as curs:
        curs.execute(f'create database {test_dbname}')
    try:
        test_url = unparse_url({**dbinfo, 'dbname': test_dbname})
        with manager.push_url(test_url):
            with cursor(test_url) as curs:
                yield curs
    finally:
        with cursor() as curs:
            curs.execute(f'drop database {test_dbname}')



from . import to_class


def to_repr(tag, **attrs):
    tag = to_repr_tag(tag)
    attrs_str = to_repr_attrs(**attrs)
    return f'<{tag}{prespace(attrs_str)}>'


def to_repr_tag(tag):
    if not isinstance(tag, str):
        tag = to_class(tag).__name__
    return tag


def to_repr_attrs(**attrs):
    return ', '.join([
        f'{name}={repr(value)}'
        for name, value in attrs.items()
    ])


def prespace(string):
    return (' ' if string else '') + string



from termcolor import colored

from ...utils.feedback import feedback
from .constants import POLECAT_ADMINISTRATOR, POLECAT_BASE, PROJECT_REGISTRY
from .exceptions import EntityExists
from .operations import (add_user_to_group, attach_bucket_policy,
                         attach_group_policy, create_bucket, create_group,
                         create_policy, create_user, find_policy,
                         get_parameter, set_parameter)
from .policy import administrator_policy, base_policy, bucket_policy
from .utils import aws_client


@feedback
def initialise(bucket, feedback):
    iam = aws_client('iam')
    s3 = aws_client('s3')
    ssm = aws_client('ssm')
    with feedback(f'Create bucket {colored(bucket, "cyan")}'):
        create_bucket(bucket, s3=s3)
    with feedback('Attach bucket policy'):
        attach_bucket_policy(bucket, bucket_policy, s3=s3)
    with feedback('Create base policy'):
        create_base_policy(iam=iam)
    with feedback('Create administrator policy'):
        create_administrator_policy(iam=iam)
    with feedback('Create administrator group'):
        create_administrator_group(iam=iam)
    with feedback('Attach administrator policy'):
        attach_administrator_policies(iam=iam)
    with feedback('Create administrator user'):
        create_administrator_user(iam=iam)
    with feedback('Attach administrator group'):
        attach_administrator_group(iam=iam)
    with feedback('Create project registry'):
        create_project_registry(ssm=ssm)


def create_base_policy(iam=None):
    create_policy(POLECAT_BASE, base_policy, iam=iam)


def create_administrator_policy(iam=None):
    create_policy(POLECAT_ADMINISTRATOR, administrator_policy)


def create_administrator_group(iam=None):
    create_group(POLECAT_ADMINISTRATOR, iam=iam)


def attach_administrator_policies(iam=None):
    attach_group_policy(
        POLECAT_ADMINISTRATOR,
        find_policy(POLECAT_BASE).arn
    )
    attach_group_policy(
        POLECAT_ADMINISTRATOR,
        find_policy(POLECAT_ADMINISTRATOR).arn
    )


def create_administrator_user(iam=None):
    access_key = create_user(POLECAT_ADMINISTRATOR, iam=iam)
    if access_key:
        # TODO: Convert
        # console.info('\nA new administrator user has been created. Please')
        # console.info('access Polecat using the following keys:\n')
        # console.info(`  AWS_ACCESS_KEY_ID = ${adminUser.AccessKeyId}`)
        # console.info(`  AWS_ACCESS_KEY_ID = ${adminUser.SecretAccessKey}\n`)
        pass


def attach_administrator_group(iam=None):
    add_user_to_group(POLECAT_ADMINISTRATOR, POLECAT_ADMINISTRATOR)


def create_project_registry(ssm=None):
    if get_parameter(PROJECT_REGISTRY, ssm=ssm) is not None:
        raise EntityExists
    set_parameter(PROJECT_REGISTRY, ['_'], ssm=ssm)



from psycopg2.sql import SQL, Identifier

from ..query import Q
from ..query import query as query_module
from ..schema.role import Role
from .delete_strategy import DeleteStrategy
from .expression.alias import Alias
from .expression.as_ import As
from .expression.cte import CTE
from .expression.expression import Expression
from .expression.json import JSON
from .expression.multi import Multi
from .expression.raw import RawSQL
from .expression.select import Select
from .expression.string_agg import StringAgg
from .expression.subquery import Subquery
from .expression.union import Union
from .expression.variable import LocalRole, LocalVariable
from .expression.where import Where
from .insert_if_missing_strategy import InsertIfMissingStrategy
from .insert_strategy import InsertStrategy
from .select_strategy import SelectStrategy
from .update_strategy import UpdateStrategy


class Strategy:
    def __init__(self):
        self.select_strategy = SelectStrategy(self)
        self.insert_strategy = InsertStrategy(self)
        self.insert_if_missing_strategy = InsertIfMissingStrategy(self)
        self.update_strategy = UpdateStrategy(self)
        self.delete_strategy = DeleteStrategy(self)

    def parse(self, queryable_or_builder):
        self.cte = CTE()
        self.query_alias_map = {}
        self.chained_relation_counter = 0
        expr = self.parse_queryable_or_builder(queryable_or_builder)
        expr = self.wrap_final_expression(expr)
        self.cte.set_final_expression(expr)
        self.push_selection_to_relations()
        return self.wrap_with_session(queryable_or_builder)

    def parse_queryable_or_builder(self, queryable_or_builder):
        if isinstance(queryable_or_builder, Q):
            return self.parse_builder(queryable_or_builder)
        else:
            return self.create_expression_from_query(queryable_or_builder)

    def parse_builder(self, builder):
        for branch in builder.iter_branches():
            self.parse_query_branch(branch)
        return self.create_expression_from_query(builder.queryable)

    def parse_query_branch(self, query):
        alias = self.get_or_create_query_alias(query)
        return alias

    def get_or_create_query_alias(self, query):
        alias = self.get_query_alias(query)
        if alias is None:
            alias = self.create_query_alias(query)
        return alias

    def get_query_alias(self, query):
        query_id = id(query)
        return self.query_alias_map.get(query_id)

    def create_query_alias(self, query):
        expression = self.create_expression_from_query(query)
        alias = self.create_alias(expression)
        query_id = id(query)
        self.query_alias_map[query_id] = alias
        return alias

    def create_alias(self, expression):
        return Alias(self.cte.append(expression))

    def create_expression_from_query(self, query):
        alias = self.get_query_alias(query)
        if alias is not None:
            return alias
        if isinstance(query, query_module.Select):
            expr = self.create_select(query)
            # TODO: Update must come before Insert, as it's
            # inherited. Probably need a better way of deciding which is
            # which. Perhaps a visitor pattern?
        elif isinstance(query, query_module.Update):
            expr = self.create_update(query)
        elif isinstance(query, query_module.InsertIfMissing):
            expr = self.create_insert_if_missing(query)
        elif isinstance(query, query_module.Insert):
            expr = self.create_insert(query)
        elif isinstance(query, query_module.Delete):
            expr = self.create_delete(query)
        elif isinstance(query, query_module.Filter):
            expr = self.create_filter(query)
        elif isinstance(query, query_module.Join):
            expr = self.create_join(query)
        elif isinstance(query, query_module.Common):
            expr = self.create_common(query)
        elif isinstance(query, Expression):
            expr = query
        else:
            raise TypeError(f'Unknown query: {query}')
        return expr

    def create_select(self, query):
        return self.select_strategy.parse_query(query)

    def create_insert(self, query):
        return self.insert_strategy.parse_query(query)

    def create_insert_if_missing(self, query):
        return self.insert_if_missing_strategy.parse_query(query)

    def create_update(self, query):
        return self.update_strategy.parse_query(query)

    def create_delete(self, query):
        return self.delete_strategy.parse_query(query)

    def create_filter(self, query):
        # TODO: Should maybe factor this out into a filter
        # strategy?
        expr = query.expression or Where(**query.options)
        return Select(
            self.parse_chained_relation(query.source),
            where=expr
        )

    def create_join(self, query):
        # TODO: Currently assumes parent is a select?
        rel = self.parse_chained_relation(query.source)
        select = rel.expression.expression
        column_name = select.columns[0]
        return Select(
            rel,
            As(
                StringAgg(
                    RawSQL(
                        SQL('{}.{}').format(
                            Identifier(rel.alias),
                            Identifier(column_name)
                        )
                    ),
                    separator=query.separator
                ),
                column_name
            )
        )

    def create_common(self, query):
        for subquery in query.subqueries[:-1]:
            expr = self.parse_queryable_or_builder(subquery)
            self.cte.append(expr)
        return self.parse_queryable_or_builder(query.subqueries[-1])

    def parse_chained_relation(self, relation):
        # TODO: I'm not too happy about the type conditional here.
        if isinstance(relation, (query_module.Insert, query_module.Update)):
            return self.get_or_create_query_alias(relation)
        elif isinstance(relation, query_module.Filter):
            counter = self.chained_relation_counter
            self.chained_relation_counter += 1
            return As(
                Subquery(
                    self.create_filter(relation)
                ),
                f'r{counter}'
            )
        elif isinstance(relation, query_module.Select):
            counter = self.chained_relation_counter
            self.chained_relation_counter += 1
            return As(
                Subquery(
                    self.create_select(relation)
                ),
                f'r{counter}'
            )
        else:
            return relation

    def wrap_final_expression(self, expression):
        if isinstance(expression, Select):
            expression = JSON(Subquery(expression))
        elif isinstance(expression, Alias):
            expression = JSON(expression)
        elif isinstance(expression, Union) or expression.returning:
            expression = JSON(Subquery(Select(Alias(self.cte.append(expression)))))
        return expression

    def push_selection_to_relations(self):
        for expr in self.cte.iter_expressions():
            expr.push_selection()

    def wrap_with_session(self, builder):
        session = getattr(builder, 'session', None)
        if not session or session.is_empty():
            return self.cte
        local_expressions = []
        if session.role:
            # TODO: Need to type this.
            assert isinstance(session.role, Role)
            local_expressions.append(LocalRole(session.role.dbname))
        for key, value in session.variables.items():
            local_expressions.append(LocalVariable(key, value))
        local_expressions.append(self.cte)
        return Multi(*local_expressions)



bucket_policy = '''{
    "Version": "2008-10-17",
    "Statement": [
        {
            "Sid": "AllowPublicRead",
            "Effect": "Allow",
            "Principal": "*",
            "Action": "s3:GetObject",
            "Resource": "arn:aws:s3:::$bucket/projects/*/bundle/*"
        },
        {
            "Effect": "Allow",
            "Principal": {
                "Service": "serverlessrepo.amazonaws.com"
            },
            "Action": "s3:GetObject",
            "Resource": "arn:aws:s3:::$bucket/projects/*/code/*"
        }
    ]
}
'''

base_policy = '''{
    "Version": "2012-10-17",
    "Statement": [
        {
            "Sid": "VisualEditor0",
            "Effect": "Allow",
            "Action": [
                "rds:*",
                "logs:*",
                "route53:ListHostedZonesByName",
                "route53:GetHostedZone",
                "route53:ChangeResourceRecordSets",
                "route53:GetChangeRequest",
                "route53:ListResourceRecordSets",
                "route53:GetChangeRequest",
                "route53:GetChange",
                "cloudformation:DescribeStackEvents",
                "acm:GetCertificate",
                "cloudformation:UpdateStack",
                "acm:ListCertificates",
                "cloudformation:ListStacks",
                "apigateway:*",
                "route53:ListHostedZones",
                "cloudformation:DescribeStackResources",
                "cloudformation:DescribeStacks",
                "cloudformation:CreateStack",
                "cloudformation:GetTemplate",
                "cloudformation:DeleteStack",
                "lambda:*",
                "iam:ListUsers",
                "iam:ListGroups",
                "iam:GetUser",
                "iam:GetRole",
                "iam:PassRole",
                "iam:CreateRole",
                "iam:DeleteRole",
                "iam:PutRolePolicy",
                "iam:DetachRolePolicy",
                "iam:DeleteRolePolicy",
                "iam:AttachRolePolicy",
                "cloudformation:ValidateTemplate",
                "cloudfront:UpdateDistribution"
            ],
            "Resource": "*"
        },
        {
            "Sid": "VisualEditor1",
            "Effect": "Allow",
            "Action": [
                "ssm:GetParameter"
            ],
            "Resource": "arn:aws:ssm:*:*:parameter/polecat/projectRegistry"
        }
    ]
}
'''

administrator_policy = '''{
    "Version": "2012-10-17",
    "Statement": [
        {
            "Sid": "VisualEditor0",
            "Effect": "Allow",
            "Action": [
                "rds:*",
                "iam:DeleteGroup",
                "iam:ListPolicies",
                "iam:AttachGroupPolicy",
                "iam:AddUserToGroup",
                "iam:CreateAccessKey",
                "iam:PutGroupPolicy",
                "iam:DeleteRolePolicy",
                "iam:ListUsers",
                "iam:ListGroups",
                "iam:GetUser",
                "iam:ListUsers",
                "iam:ListGroups",
                "iam:GetUser",
                "iam:GetRole",
                "iam:PassRole",
                "iam:CreateRole",
                "iam:DeleteRole",
                "iam:PutRolePolicy",
                "iam:DetachRolePolicy",
                "iam:DeleteRolePolicy",
                "iam:AttachRolePolicy",
                "iam:CreateUser",
                "iam:GetGroup",
                "iam:CreateGroup",
                "iam:DeleteUser",
                "iam:CreatePolicy",
                "iam:ListGroupsForUser",
                "iam:DetachGroupPolicy",
                "iam:DeletePolicy",
                "logs:*",
                "s3:CreateBucket",
                "s3:ListBucket",
                "s3:DeleteObject",
                "s3:DeleteBucket",
                "s3:PutObject",
                "s3:GetObject",
                "s3:PutBucketPolicy",
                "route53:ListHostedZonesByName",
                "cloudformation:DescribeStackEvents",
                "acm:GetCertificate",
                "acm:ListCertificates",
                "acm:ListTagsForCertificate",
                "acm:RequestCertificate",
                "acm:DescribeCertificate",
                "acm:AddTagsToCertificate",
                "cloudformation:UpdateStack",
                "route53:DeleteHostedZone",
                "cloudformation:ListStacks",
                "route53:CreateHostedZone",
                "apigateway:*",
                "route53:ListHostedZones",
                "cloudformation:DescribeStackResources",
                "cloudformation:DescribeStacks",
                "cloudformation:CreateStack",
                "cloudformation:GetTemplate",
                "cloudformation:DeleteStack",
                "lambda:*",
                "ce:*",
                "cloudformation:ValidateTemplate"
            ],
            "Resource": "*"
        },
        {
            "Sid": "VisualEditor1",
            "Effect": "Allow",
            "Action": [
                "ssm:PutParameter",
                "ssm:DeleteParameter",
                "ssm:DeleteParameters",
                "ssm:GetParametersByPath",
                "ssm:GetParameters",
                "ssm:GetParameter"
            ],
            "Resource": "arn:aws:ssm:*:*:parameter/polecat/*"
        }
    ]
}
'''

project_policy = '''{
    "Version": "2012-10-17",
    "Statement": [
        {
            "Sid": "VisualEditor0",
            "Effect": "Allow",
            "Action": [
                "rds:*",
                "logs:*",
                "route53:ListHostedZonesByName",
                "route53:GetHostedZone",
                "route53:ChangeResourceRecordSets",
                "route53:GetChangeRequest",
                "route53:ListResourceRecordSets",
                "route53:GetChangeRequest",
                "route53:GetChange",
                "cloudformation:DescribeStackEvents",
                "acm:GetCertificate",
                "cloudformation:UpdateStack",
                "acm:ListCertificates",
                "cloudformation:ListStacks",
                "apigateway:*",
                "route53:ListHostedZones",
                "cloudformation:DescribeStackResources",
                "cloudformation:DescribeStacks",
                "cloudformation:CreateStack",
                "cloudformation:GetTemplate",
                "cloudformation:DeleteStack",
                "lambda:*",
                "iam:ListUsers",
                "iam:ListGroups",
                "iam:GetUser",
                "iam:GetRole",
                "iam:PassRole",
                "iam:CreateRole",
                "iam:DeleteRole",
                "iam:PutRolePolicy",
                "iam:DetachRolePolicy",
                "iam:DeleteRolePolicy",
                "iam:AttachRolePolicy",
                "cloudformation:ValidateTemplate"
            ],
            "Resource": "*"
        },
        {
            "Sid": "VisualEditor1",
            "Effect": "Allow",
            "Action": [
                "ssm:PutParameter",
                "ssm:DeleteParameter",
                "ssm:DeleteParameters",
                "ssm:GetParametersByPath",
                "ssm:GetParameters",
                "ssm:GetParameter"
            ],
            "Resource": "arn:aws:ssm:*:*:parameter/polecat/projects/$project/*"
        },
        {
            "Sid": "VisualEditor2",
            "Effect": "Allow",
            "Action": [
                "ssm:GetParameter"
            ],
            "Resource": "arn:aws:ssm:*:*:parameter/polecat/projectRegistry"
        },
        {
            "Sid": "VisualEditor3",
            "Effect": "Allow",
            "Action": [
                "s3:ListBucket",
                "s3:DeleteObject",
                "s3:PutObject",
                "s3:GetObject"
            ],
            "Resource": [
                "arn:aws:s3:::$bucket/projects/$project",
                "arn:aws:s3:::$bucket/projects/$project/*"
            ]
        }
    ]
}
'''



from polecat.db.schema import CASCADE  # noqa
from polecat.db.schema.utils import Auto  # noqa

from .blueprint import Blueprint  # noqa
from .defaults import default_blueprint  # noqa
from .field import *  # noqa
from .model import Access, Model, Mutation, Query, Role, Type  # noqa
from .session import session  # noqa



__all__ = ('Schema',)


class Schema:
    def __init__(self, tables=None, roles=None, access=None):
        self.tables = tables or []
        self.tables_by_name = {
            t.name: t
            for t in self.tables
        }
        self.roles = roles or []
        self.roles_by_name = {
            r.name: r
            for r in self.roles
        }
        self.access = access or []
        self.access_by_entity = {
            a.entity.signature: a
            for a in self.access
        }

    def add_table(self, *tables):
        for table in tables:
            self.tables.append(table)
            self.tables_by_name[table.name] = table

    def add_role(self, *roles):
        for role in roles:
            self.roles.append(role)
            self.roles_by_name[role.name] = role

    def add_access(self, *accesses):
        for access in accesses:
            self.access.append(access)
            self.access_by_entity[access.entity.signature] = access

    def remove_access(self, *accesses):
        for access in accesses:
            for index, existing in enumerate(self.access):
                if existing.signature == access.signature:
                    self.access.pop(index)
                    del self.access_by_entity[existing.signature]
                    return

    def bind(self):
        for table in self.tables:
            self.bind_table(table)
        for role in self.roles:
            self.bind_role(role)
        for access in self.access:
            self.bind_access(access)

    def bind_table(self, table):
        table.bind(self)

    def bind_role(self, role):
        role.bind(self)

    def bind_access(self, access):
        access.bind(self)

    def get_table_by_name(self, name):
        return self.tables_by_name[name]

    def get_role_by_name(self, name):
        return self.roles_by_name[name]

    def get_access_by_entity(self, entity):
        return self.access_by_entity[entity.signature]

    def merge_access(self):
        all_access = {}
        for access in self.access:
            all_access[access.signature] = access
        return tuple(all_access.values())



import asyncio
from jwt import InvalidSignatureError

import ujson
from graphql_server.error import HttpQueryError

from ...project.project import load_project
from .event import LambdaEvent


class LambdaServer:
    def __init__(self, project=None):
        self.project = project or load_project()
        self.project.prepare()

    def handle(self, event, context=None):
        event = LambdaEvent(event)
        if event.request.method == 'OPTIONS':
            return self.handle_cors(event)
        loop = asyncio.get_event_loop()
        try:
            result = loop.run_until_complete(
                self.project.handle_event(event)
            )
        except InvalidSignatureError:
            result = (
                {'errors': ['Unauthorized']},
                401
            )
        except HttpQueryError as e:
            result = (
                {'errors': [e.message]},
                e.status_code
            )
        return self.encode_result(event, result)

    def handle_cors(self, event):
        # TODO: Need to make this more restrictive.
        return {
            'statusCode': '200',
            'body': '',
            'headers': {
                'Access-Control-Allow-Origin': '*',
                'Access-Control-Allow-Methods': 'GET,PUT,POST,PATCH,OPTIONS',
                'Access-Control-Allow-Headers': 'Content-Type,Authorization',
                'Access-Control-Max-Age': '86400'
            }
        }

    def encode_result(self, event, result):
        # TODO: I think separating admin from server makes more sense.
        # I'll come back here and split this out.
        if event.is_admin():
            return result
        body = result[0]
        headers = {
            # TODO: More restrictive.
            'Access-Control-Allow-Origin': '*',
        }
        content_type = result[2] if len(result) > 2 else None
        if content_type != 'text/html':
            body = ujson.dumps(body)
        elif content_type is not None:
            headers['content-type'] = 'text/html'
        return {
            'statusCode': result[1],
            'body': body,
            'headers': headers
        }



from graphql import GraphQLError
from graphql.type import GraphQLList
from polecat.model.db import Q, S
from polecat.model.resolver import APIContext

from ..utils.exceptions import traceback
from .field import RelatedField
from .input import Input, parse_id


class GraphQLAPIContext(APIContext):
    def __init__(self, root, info, **kwargs):
        super().__init__()
        self.root = root
        self.info = info
        self.kwargs = kwargs
        self.event = self.graphql_context.get('event')
        self.session = self.graphql_context.get('session')

    def parse_argument(self, name):
        return self.kwargs.get(name)

    def parse_input(self):
        return Input(self.input_type, self.kwargs.get('input', {})).change

    def raw_input(self):
        return self.kwargs['input']

    @property
    def field_name(self):
        return self.info.field_name

    @property
    def field(self):
        # TODO: Cache.
        return self.info.parent_type.fields[self.field_name]

    @property
    def return_type(self):
        return self.info.return_type

    @property
    def root_node(self):
        return self.info.field_nodes[0]

    @property
    def graphql_context(self):
        return self.info.context or {}

    @property
    def model_class(self):
        try:
            return self.return_type._model
        except AttributeError:
            return self.return_type.of_type._model

    @property
    def query(self):
        return self.field._query

    @property
    def mutation(self):
        return self.field._mutation

    @property
    def input_type(self):
        return self.field.args['input'].type

    def parse_id(self):
        try:
            id = parse_id(self.kwargs['id'])
        except KeyError:
            id = None
        return id

    def get_selector(self):
        return get_selector_from_node(self.return_type, self.root_node)


def resolve_all_query(obj, info, **kwargs):
    with traceback():
        ctx = GraphQLAPIContext(obj, info, **kwargs)
        return ctx.model_class.Meta.all_resolver_manager(ctx)


def resolve_get_query(obj, info, **kwargs):
    with traceback():
        ctx = GraphQLAPIContext(obj, info, **kwargs)
        return ctx.model_class.Meta.get_resolver_manager(ctx)


def resolve_create_mutation(obj, info, **kwargs):
    with traceback():
        ctx = GraphQLAPIContext(obj, info, **kwargs)
        return ctx.model_class.Meta.create_resolver_manager(ctx)


def resolve_update_mutation(obj, info, **kwargs):
    with traceback():
        ctx = GraphQLAPIContext(obj, info, **kwargs)
        return ctx.model_class.Meta.update_resolver_manager(ctx)


def resolve_update_or_create_mutation(obj, info, **kwargs):
    with traceback():
        ctx = GraphQLAPIContext(obj, info, **kwargs)
        return ctx.model_class.Meta.update_or_create_resolver_manager(ctx)


def resolve_delete_mutation(obj, info, **kwargs):
    with traceback():
        ctx = GraphQLAPIContext(obj, info, **kwargs)
        return ctx.model_class.Meta.delete_resolver_manager(ctx)
    # return_type = info.return_type
    # id = parse_id(kwargs['input']['id'])
    # model_class = return_type._model
    # model = model_class(id=id)
    # options = info.context or {}
    # Q(model).delete().execute(**options)
    # return {
    #     'id': id
    # }


def build_all_query(graphql_type, node, query=None, options=None):
    model = graphql_type._model
    if not query:
        query = Q(model, **(options or {}))
    return query.select(get_selector_from_node(graphql_type, node))


def get_selector_from_node(graphql_type, node):
    # TODO: I'd like this to be bundled up in the GraphQL type itself,
    # as in we should call "graphql_type.polecat_fields" to get our
    # type. This means altering the instantiated GQLType during schema
    # construction.
    if isinstance(graphql_type, GraphQLList):
        graphql_type = graphql_type.of_type
    fields = []
    lookups = {}
    for field_node in node.selection_set.selections:
        cc_field_name = field_node.name.value
        try:
            graphql_field = graphql_type.fields[cc_field_name]
        except KeyError:
            raise GraphQLError(f'No field "{cc_field_name}" on type "{graphql_type}"')
        model_field = graphql_field._field.model_field
        if isinstance(graphql_field._field, RelatedField):
            lookups[model_field.name] = get_selector_from_node(
                graphql_field.type,
                field_node
            )
        else:
            fields.append(model_field.name)
    return S(*fields, **lookups)


def resolve_query(obj, info, **kwargs):
    # graphql_field = info.parent_type.fields[info.field_name]
    # query = graphql_field._query
    # # TODO: Need to pass the context.
    # return query.resolve(**kwargs)
    with traceback():
        ctx = GraphQLAPIContext(obj, info, **kwargs)
        return ctx.query.Meta.resolver_manager(ctx)


def resolve_mutation(obj, info, **kwargs):
    # node = info.field_nodes[0]
    # graphql_field = info.parent_type.fields[info.field_name]
    # return_type = graphql_field.type
    # options = info.context or {}
    # mutation = graphql_field._mutation(
    #     selector=get_selector_from_node(return_type, node),
    #     **options
    # )
    # api_context = GraphQLAPIContext(obj, info, **kwargs)
    # input = api_context.parse_input()
    # return mutation.resolve(**input)
    with traceback():
        ctx = GraphQLAPIContext(obj, info, **kwargs)
        return ctx.mutation.resolver_manager(ctx)



from psycopg2.sql import Identifier

from ...utils.repr import to_repr
from .column import MutableColumn
from .entity import Entity

__all__ = ('Table',)


class Table(Entity):
    # TODO: These are here to support the Queryable interface. Bad
    # design, need to encapsulate these somewhere else.
    selectable = True
    mutatable = True

    def __init__(self, name, columns=None, checks=None, uniques=None,
                 access=None, indexes=None, policies=None, app=None):
        # TODO: Unknown-unknown; cannot rename "name" or
        # polecat.db.query.selection.Selection will break. It expects
        # that if the column name passed in is not a string it has a
        # `.name` attribute.
        self.app = app
        self._name = name  # TODO: Not sure how I feel about this.
        self.columns = columns or []
        self.checks = checks or []
        self.uniques = uniques or []
        self.access = access
        self.indexes = indexes or []
        self.policies = policies or []
        self.C = type('Columns', (), {})  # TODO
        self.schema = None

    def __repr__(self):
        return to_repr(
            self,
            app=self.app,
            name=self.name
        )

    @property
    def signature(self):
        return (Table, getattr(self.app, 'name', self.app), self.name)

    @property
    def name(self):
        name = self._name
        if self.app:
            name = f'{self.app.name.lower()}_{name}'
        return name

    def has_changed(self, other):
        return (
            getattr(self.app, 'name', self.app) != getattr(other.app, 'name', other.app) or
            self.name != other.name or
            self.checks != other.checks or
            self.uniques != other.uniques or
            self.have_columns_changed(other)
        )

    def have_columns_changed(self, other):
        # TODO: I don't think this is sufficient. I think I need to be
        # running "has_changed" on the columns in order to detect
        # variations to columns.
        from_columns = {
            column.name: column
            for column in self.iter_mutable_columns()
        }
        to_columns = {
            column.name: column
            for column in other.iter_mutable_columns()
        }
        return from_columns != to_columns

    # TODO: This is for making SQL expressions. It would be nice to
    # not have this here, as it's bleeding information between
    # abstractions.
    @property
    def alias(self):
        return self.name

    @property
    def root_relation(self):
        return self

    def add_column(self, column):
        self.columns.append(column)
        self.bind_column(column)

    def add_index(self, index):
        self.indexes.append(index)
        self.bind_index(index)

    def add_policy(self, policy):
        self.policies.append(policy)
        self.bind_policy(policy)

    # TODO: This is for making SQL expressions. It would be nice to
    # not have this here, as it's bleeding information between
    # abstractions.
    def to_sql(self):
        return Identifier(self.name), ()

    def bind(self, schema):
        if self.schema is None:
            self.schema = schema
            self.bind_all_columns()
            self.bind_all_indexes()
            self.bind_all_policies()

    def bind_all_columns(self):
        for column in self.columns:
            self.bind_column(column)

    def bind_column(self, column):
        setattr(self.C, column.name, column)
        column.bind(self)

    def bind_all_indexes(self):
        for index in self.indexes:
            self.bind_index(index)

    def bind_all_policies(self):
        for policy in self.policies:
            self.bind_policy(policy)

    def bind_index(self, index):
        index.bind(self)

    def bind_policy(self, policy):
        policy.bind(self)

    def is_rls_enabled(self):
        # TODO: How to know really?
        return False

    def has_column(self, name):
        return hasattr(self.C, name)

    def get_column(self, name):
        try:
            return getattr(self.C, name)
        except AttributeError:
            raise KeyError(
                f'Column "{name}" does not exist on table "{self.name}"'
            )

    def iter_mutable_columns(self):
        for column in self.columns:
            if isinstance(column, MutableColumn):
                yield column

    def iter_column_names(self):
        for column in self.columns:
            yield column.name

    def get_subrelation(self, name):
        column = self.get_column(name)
        return column.related_table

    def push_selection(self, selection=None):
        pass



import sys
from functools import wraps

from halo import Halo
from polecat.core.config import default_config
from polecat_feedback.decorators import feedback
from polecat_feedback.renderer import Renderer
from termcolor import colored

from ..utils.feedback import Feedback


def cli_feedback(title):
    def outer(func):
        @wraps(func)
        def inner(*args, **kwargs):
            ctx = args[0]
            renderer = Renderer(max_width=88)
            with feedback(title, app_name='Polecat', version='v0.0.8', renderer=renderer) as fb:
                path = ctx.obj.get('config_path')
                if path:
                    # TODO: Should be hiding this.
                    path = fb.renderer.t.bright_white(path)
                    fb.add_notice(f'Loaded configuration from {path}')
                try:
                    return func(*args, feedback=fb, **kwargs)
                except Exception as e:
                    # TODO: Add error handling.
                    raise
        return inner
    return outer


class HaloFeedback(Feedback):
    def __init__(self, message=None):
        self.spinner = Halo(text=message or '', spinner='dots')
        if message and not default_config.debug:
            self.spinner.start()

    def update_message(self, message):
        super().update_message(message)
        if not self.spinner._spinner_id and not default_config.debug:
            self.spinner.start()
        self.spinner.text = (message + ' ...') if self.message else ''

    def succeeded(self):
        self.spinner.text = self.message
        self.spinner.succeed()

    def errored(self, error):
        # self.spinner.text = str(error) if error else self.message
        self.spinner.text = f'{self.message} ... {colored(str(error), "red")}'
        self.spinner.fail()
        sys.exit(1)

    def warning(self, warning):
        self.spinner.text = f'{self.message} ... {colored(str(warning), "yellow")}'
        self.spinner.warn()

    def info(self, message):
        self.spinner.info(message)

    def exists(self, error):
        self.warning('exists')



from functools import wraps

from ..utils.decorator import decorator_with_args
from . import connection


@decorator_with_args
def dbcursor(func, autocommit=True):
    @wraps(func)
    def inner(*args, cursor=None, **kwargs):
        if cursor is None:
            with connection.cursor(autocommit=autocommit) as curs:
                return func(*args, cursor=curs, **kwargs)
        else:
            # TODO: Make sure autocommit matches?
            return func(*args, cursor=cursor, **kwargs)
    return inner



from polecat.utils import add_attribute, to_list
from polecat.utils.stringcase import camelcase, snakecase

from .defaults import default_blueprint
from .field import Field, SerialField
from .omit import NONE


class RegisteredType(type):
    BASE_TYPE_NAME = None

    def build_type(metaclass, name, bases, attrs):
        cls = super().__new__(metaclass, name, bases, attrs)
        if metaclass.is_sub_type(metaclass, name, bases):
            metaclass.register_type(metaclass, cls)
        return cls

    def is_sub_type(metaclass, name, bases):
        return bases and name != metaclass.BASE_TYPE_NAME

    def register_type(metaclass, cls):
        pass


class ConstructableMetaType(RegisteredType):
    def build_type(metaclass, name, bases, attrs):
        is_sub = metaclass.is_sub_type(metaclass, name, bases)
        if is_sub:
            meta = metaclass.build_meta(
                metaclass, name, attrs, attrs.get('Meta')
            )
            if meta is not None:
                attrs['Meta'] = meta
            cls = super().__new__(metaclass, name, bases, {
                k: v
                for k, v in attrs.items()
                if k not in attrs['Meta'].fields
            })
        else:
            cls = super().__new__(metaclass, name, bases, attrs)
        if is_sub:
            metaclass.register_type(metaclass, cls)
        return cls

    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        meta = getattr(cls, 'Meta', None)
        if meta:
            # Must first dump all fields as any related field that
            # references "self" will cause an added reverse field to
            # be added.
            all_fields = tuple(meta.fields.values())
            for field in all_fields:
                field.prepare(cls)

    def build_meta(metaclass, name, attrs, meta):
        return None


class TypeMetaclass(ConstructableMetaType):
    BASE_TYPE_NAME = 'Type'

    def __new__(meta, name, bases, attrs):
        return meta.build_type(meta, name, bases, attrs)

    def build_meta(metaclass, name, attrs, meta):
        return make_type_meta(name, attrs, meta)

    def register_type(metaclass, cls):
        default_blueprint.add_type(cls)


class ModelMetaclass(ConstructableMetaType):
    BASE_TYPE_NAME = 'Model'

    def __new__(metaclass, name, bases, attrs):
        return metaclass.build_type(metaclass, name, bases, attrs)

    def build_meta(metaclass, name, attrs, meta):
        return make_model_meta(name, attrs, meta)

    def register_type(metaclass, cls):
        default_blueprint.add_model(cls)


class RoleMetaclass(RegisteredType):
    BASE_TYPE_NAME = 'Role'

    def __new__(metaclass, name, bases, attrs):
        is_sub = metaclass.is_sub_type(metaclass, name, bases)
        if is_sub:
            attrs['Meta'] = make_role_meta(
                name, bases, attrs, attrs.get('Meta')
            )
        cls = super().__new__(metaclass, name, bases, make_role_attrs(attrs))
        if is_sub:
            metaclass.register_type(metaclass, cls)
        return cls

    def register_type(metaclass, cls):
        default_blueprint.add_role(cls)


class AccessMetaclass(RegisteredType):
    BASE_TYPE_NAME = 'Access'

    def __new__(metaclass, name, bases, attrs):
        is_sub = metaclass.is_sub_type(metaclass, name, bases)
        cls = super().__new__(metaclass, name, bases, make_access_attrs(attrs))
        if is_sub:
            metaclass.register_type(metaclass, cls)
        return cls

    def register_type(metaclass, cls):
        default_blueprint.add_access(cls)


class QueryMetaclass(RegisteredType):
    BASE_TYPE_NAME = 'Query'

    def __new__(metaclass, name, bases, attrs):
        is_sub = metaclass.is_sub_type(metaclass, name, bases)
        if is_sub:
            attrs['Meta'] = make_query_meta(
                name, bases, attrs, attrs.get('Meta')
            )
        cls = super().__new__(metaclass, name, bases, make_query_attrs(attrs))
        if is_sub:
            metaclass.register_type(metaclass, cls)
        return cls

    def register_type(metaclass, cls):
        default_blueprint.add_query(cls)


class MutationMetaclass(RegisteredType):
    def __new__(metaclass, name, bases, attrs):
        is_sub = metaclass.is_sub_type(metaclass, name, bases)
        cls = super().__new__(
            metaclass, name, bases, make_mutation_attrs(name, attrs)
        )
        if is_sub:
            metaclass.register_type(metaclass, cls)
        return cls

    def register_type(metaclass, cls):
        default_blueprint.add_mutation(cls)


# TODO: So similar to models.
def make_type_meta(name, attrs, meta):
    fields = {
        f.name: f
        for f in get_type_fields(attrs)
    }
    return type('Meta', (), {
        'options': meta.__dict__ if meta else {},
        'name': name,
        'fields': fields,
        'cc_fields': make_cc_fields(fields),
        'plural': get_plural(name, meta),
        'input': getattr(meta, 'input', False) if meta else False,  # TODO: Ugly
        'omit': getattr(meta, 'omit', NONE) if meta else NONE  # TODO: Ugly
    })


def make_model_meta(name, attrs, meta):
    # TODO: Ugh.
    from .resolver import (
        CreateResolver, ResolverList, CreateResolverManager,
        UpdateResolverManager, UpdateResolver, UpdateOrCreateResolverManager,
        AllResolver, AllResolverManager, GetResolver, GetResolverManager,
        DeleteResolver, DeleteResolverManager
    )
    fields = {
        f.name: f
        for f in get_model_fields(attrs)
    }
    query_resolvers = to_list(getattr(meta, 'query_resolvers', []))
    mutation_resolvers = to_list(getattr(meta, 'mutation_resolvers', []))
    create_resolvers = to_list(getattr(meta, 'create_resolvers', []))
    return type('Meta', (), {
        'options': meta.__dict__ if meta else {},
        'app': getattr(meta, 'app', None),
        'name': name,
        'table_name': name.lower(),
        'fields': fields,
        'cc_fields': make_cc_fields(fields),
        'plural': get_plural(name, meta),
        'uniques': getattr(meta, 'uniques', ()) if meta else (),
        'checks': getattr(meta, 'checks', ()) if meta else (),
        'omit': getattr(meta, 'omit', NONE) if meta else NONE,  # TODO: Duplicate of above
        'indexes': getattr(meta, 'indexes', ()),
        'policies': getattr(meta, 'policies', ()),
        'query_resolvers': ResolverList(query_resolvers),
        'all_resolvers': ResolverList(AllResolver()),
        'get_resolvers': ResolverList(GetResolver()),
        'all_resolver_manager': AllResolverManager(),
        'get_resolver_manager': GetResolverManager(),
        'mutation_resolvers': ResolverList(mutation_resolvers),
        'create_resolvers': ResolverList(create_resolvers + [CreateResolver()]),
        'update_resolvers': ResolverList(UpdateResolver()),
        'delete_resolvers': ResolverList(DeleteResolver()),
        'create_resolver_manager': CreateResolverManager(),
        'update_resolver_manager': UpdateResolverManager(),
        'update_or_create_resolver_manager': UpdateOrCreateResolverManager(),
        'delete_resolver_manager': DeleteResolverManager()
    })


def make_role_meta(name, bases, attrs, meta):
    return type('Meta', (), {
        'options': meta.__dict__ if meta else {},
        'app': None,
        'name': name,
        'role': make_role_from_name(name)
    })


def make_query_meta(name, bases, attrs, meta):
    meta_name = getattr(meta, 'name', None)
    if meta_name is not None:
        name = meta_name
    elif name[-5:] == 'Query':
        name = name[:-5]
    # TODO: Ugh.
    from .resolver import QueryResolverManager
    resolver_manager = attrs.get('resolver_manager')
    if not resolver_manager:
        resolver_manager = QueryResolverManager()
    return type('Meta', (), {
        'options': meta,
        'name': name,
        'resolvers': attrs.get('resolvers', ()),
        'resolver_manager': resolver_manager,
        'omit': getattr(meta, 'omit', NONE) if meta else NONE  # TODO: Duplicate of above
    })


def make_mutation_attrs(name, attrs):
    # TODO: Ugh.
    from .resolver import MutationResolverManager
    resolver_manager = attrs.get('resolver_manager')
    if not resolver_manager:
        resolver_manager = MutationResolverManager()
    return {
        **attrs,
        'name': attrs.get('name', name),
        'resolvers': attrs.get('resolvers', ()),
        'resolver_manager': resolver_manager,
        'omit': getattr(attrs, 'omit', NONE)
    }


def make_role_attrs(attrs):
    return {
        'app': None,
        **attrs,
        'parents': attrs.get('parents', ())
    }


def make_access_attrs(attrs):
    return {
        'app': None,
        'all': (),
        'select': (),
        'insert': (),
        'update': (),
        'delete': (),
        **attrs
    }


def make_query_attrs(attrs):
    return {
        **attrs
    }


def get_type_fields(attrs):
    for name, field in attrs.items():
        if is_field_class(field):
            yield add_attribute(
                add_attribute(field, 'name', name),
                'cc_name', camelcase(name)
            )


def get_model_fields(attrs):
    yield add_attribute(
        add_attribute(
            SerialField(primary_key=True), 'name', 'id'
        ), 'cc_name', 'id'
    )
    for field in get_type_fields(attrs):
        yield field


def make_cc_fields(fields):
    return {
        f.cc_name: f
        for f in fields.values()
    }


def get_plural(name, meta):
    return getattr(meta, 'plural', f'{name}s')


def is_field_class(value):
    try:
        return issubclass(value.__class__, Field)
    except TypeError:
        return False


def make_role_from_name(name):
    name = snakecase(name)
    if name[-5:] == '_role':
        name = name[:-5]
    return name



import time

from polecat.utils import to_list
from polecat.utils.feedback import feedback
from termcolor import colored

from .exceptions import KnownError
from .utils import aws_client


@feedback
def create_certificate(domains, name=None, feedback=None):
    domains = to_list(domains)
    acm = aws_client('acm', region='us-east-1')
    r53 = aws_client('route53')
    main_domain = domains[0]
    alternatives = domains[1:]
    with feedback(f'Creating certificate request for {colored(main_domain, "yellow")}'):
        hosted_zone_id = get_hosted_zone_id(main_domain, r53=r53)
        options = {
            'DomainName': main_domain,
            'ValidationMethod': 'DNS',
        }
        if alternatives:
            options['SubjectAlternativeNames'] = alternatives
        response = acm.request_certificate(**options)
        cert_arn = response['CertificateArn']
        acm.add_tags_to_certificate(
            CertificateArn=cert_arn,
            Tags=[{
                'Key': 'Name',
                'Value': name
            }]
        )
        attempt = 0
        while True:
            response = acm.describe_certificate(CertificateArn=cert_arn)
            options = response['Certificate']['DomainValidationOptions'][0]
            try:
                record = options['ResourceRecord']
                break
            except KeyError:
                attempt += 1
                if attempt == 3:
                    raise
                time.sleep(5)
        r53.change_resource_record_sets(
            HostedZoneId=hosted_zone_id,
            ChangeBatch={
                'Changes': [{
                    'Action': 'CREATE',
                    'ResourceRecordSet': {
                        'Name': record['Name'],
                        'Type': record['Type'],
                        'TTL': 300,
                        'ResourceRecords': [{
                            'Value': record['Value']
                        }]
                    }
                }]
            }
        )


@feedback
def wait_certificate(domain, feedback=None):
    acm = aws_client('acm', region='us-east-1')
    with feedback(f'Waiting for certificate {colored(domain, "yellow")}'):
        cert_arn = get_certificate_arn(domain)
        waiter = acm.get_waiter('certificate_validated')
        try:
            waiter.wait(CertificateArn=cert_arn)
        except:
            raise Warning('timed out (try again)')


def get_certificate_arn(domain, acm=None):
    acm = aws_client('acm', client=acm, region='us-east-1')
    paginator = acm.get_paginator('list_certificates')
    possible_certs = []
    for info in paginator.paginate():
        for cert in info['CertificateSummaryList']:
            tags = acm.list_tags_for_certificate(
                CertificateArn=cert['CertificateArn']
            )['Tags']
            name_tag = [
                t for t in tags
                if t['Key'] == 'Name' and t['Value'] == domain
            ]
            if len(name_tag) > 0:
                return cert['CertificateArn']
            elif cert['DomainName'] == domain:
                return cert['CertificateArn']
            possible_certs.append(cert['DomainName'])
    error = KnownError('unknown domain')
    error.possible_certificates = possible_certs
    raise error


def get_hosted_zone_id(domain, r53=None):
    r53 = aws_client('r53', client=r53)
    dns_name = get_dns_name_from_domain(domain)
    response = r53.list_hosted_zones_by_name(DNSName=dns_name)
    try:
        return response['HostedZones'][0]['Id']
    except (KeyError, IndexError):
        raise KnownError('unknown domain')


def get_dns_name_from_domain(domain):
    return '.'.join(domain.split('.')[-2:]) + '.'



POLECAT_BASE = 'PolecatBase'
POLECAT_ADMINISTRATOR = 'PolecatAdministrator'

PROJECT_REGISTRY = '/polecat/projectRegistry'
PROJECT_PREFIX = '/polecat/projects/{}/'
DEPLOYMENT_REGISTRY = PROJECT_PREFIX + 'deploymentRegistry'
DEPLOYMENT_PREFIX = PROJECT_PREFIX + 'deployments/{}/'
SECRET_PREFIX = DEPLOYMENT_PREFIX + 'secrets/'
SECRET_NAME = SECRET_PREFIX + '{}'
DOMAIN_PREFIX = DEPLOYMENT_PREFIX + 'domains/{}/'
CERTIFICATE = DOMAIN_PREFIX + 'certificate'
ZONE = DOMAIN_PREFIX + 'zone'

SERVER_CODE = 'projects/{}/code/{}/server.zip'
MEDIA_PREFIX = 'projects/{}/deployments/{}/media/'



from contextlib import contextmanager
from traceback import print_tb


class EntityExists(Exception):
    pass


class EntityDoesNotExist(Exception):
    pass


class KnownError(Exception):
    pass


@contextmanager
def traceback():
    try:
        yield
    except Exception as e:
        print_tb(e.__traceback__)
        raise



from stdb.core.engine import default_engine as engine

from .query import Query


class Delete(Query):
    def __init__(self, version, endpoint, id, **kwargs):
        super().__init__(version, endpoint, **kwargs)
        self.id = id

    def execute(self):
        sql = self.get_sql()
        engine.execute(sql)

    def get_sql(self):
        return (
            'DELETE FROM {} WHERE id = {};'
        ).format(
            self.endpoint['table'],
            engine.quote_literal(self.id)
        )



from .importing import import_attribute


class Proxy:
    def __init__(self, initfunc=None):
        self.__target = None
        self.__initfunc = initfunc

    def set_target(self, target):
        self.__target = target

    def get_target(self):
        return self.__target

    def __getattr__(self, name):
        if self.__target is None and self.__initfunc:
            self.__initfunc = import_attribute(self.__initfunc)
            self.__target = self.__initfunc()
        return getattr(self.__target, name)

    def __getitem__(self, *args, **kwargs):
        if self.__target is None and self.__initfunc:
            self.__initfunc = import_attribute(self.__initfunc)
            self.__target = self.__initfunc()
        return self.__target.__getitem__(*args, **kwargs)



from polecat.db.query import query as query_module

from .expression.delete import Delete
from .expression.where import Where


class DeleteStrategy:
    def __init__(self, root_strategy):
        self.root = root_strategy

    def parse_query(self, query):
        if isinstance(query.source, query_module.Filter):
            relation = query.source.source
            where = Where(**query.source.options)
        else:
            relation = query.source
            where = None
        return Delete(relation, where=where)



from .handler import Handler

index_html = '''<!DOCTYPE html>
<html>
  <head>
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <meta charset="UTF-8">
    <title>{default_title}</title>
  </head>
  <body>
    <div id="mount"></div>
    <script type="text/javascript">
      window.publicPath = 'https://s3-{region}.amazonaws.com/{bucket}/projects/{project}/bundle/{bundle_version}/'
    </script>
    <script
      type="text/javascript"
      src="https://s3-{region}.amazonaws.com/{bucket}/projects/{project}/bundle/{bundle_version}/{bundle}"
    >
    </script>
  </body>
</html>
'''


def get_index_html(bucket, project, bundle_version, region='ap-southeast-2',
                   bundle='index.js.gz', default_title=''):
    return index_html.format(
        bucket=bucket,
        project=project,
        bundle_version=bundle_version,
        region=region,
        bundle=bundle,
        default_title=default_title
    )


class IndexHandler(Handler):
    def prepare(self):
        self.index_html = self.project.get_index_html()

    def match(self, event):
        # TODO: Use regex
        return event.is_http() and \
           event.request.method == 'GET' or \
           event.request.path == '/'

    async def handle_event(self, event):
        return (self.index_html, 200, 'text/html')



import re

from jwt import decode
from polecat.core.config import default_config
from polecat.model import default_blueprint

__all__ = ('JWTMiddleware', 'RoleMiddleware')


class JWTMiddleware:
    bearer_prog = re.compile(r'bearer\s+(\S+)', re.I)

    def run(self, event):
        claims = None
        jwt = event.get_authorization_header()
        if jwt:
            match = self.bearer_prog.match(jwt)
            if match:
                jwt = match.group(1)
                claims = decode(
                    jwt,
                    default_config.jwt_secret,
                    algorithms=('HS256',)
                )
        event.claims = claims
        for key, value in (claims or {}).items():
            event.session.variables[f'claims.{key}'] = value


class RoleMiddleware(JWTMiddleware):
    def __init__(self, default_role=None):
        super().__init__()
        self.default_role = default_role

    def run(self, event):
        super().run(event)
        role = self.default_role
        claims = event.claims
        if claims and 'role' in claims:
            role = default_blueprint.roles[claims['role']]
        if role is None:
            # TODO: Better exception.
            raise Exception('No role specified')
        # TODO: Need both of these? Probs not.
        event.role = role
        event.session.role = role.Meta.dbrole if role else None



from psycopg2.sql import SQL, Identifier

from .expression import Expression


class Delete(Expression):
    def __init__(self, relation, returning=None, where=None):
        self.relation = relation
        self.returning = returning or ()
        self.where = where

    def to_sql(self):
        if self.returning:
            returning_sql = SQL(' RETURNING {}').format(
                map(Identifier, self.returning)
            )
        else:
            returning_sql = SQL('')
        where_sql, where_args = self.get_where_sql()
        return (
            SQL('DELETE FROM {}{}{}').format(
                Identifier(self.relation.name),
                returning_sql,
                where_sql
            ),
            where_args
        )

    def get_where_sql(self):
        if self.where:
            sql, args = self.where.get_sql(self.relation)
            return SQL(' WHERE {}').format(sql), args
        else:
            return SQL(''), ()

    def push_selection(self, selection=None):
        pass



from stdb.core.engine import default_engine as engine

from .coerce import coerce
from .query import Query


class BulkCreate(Query):
    def __init__(self, version, endpoint, data, ignore_duplicates=False):
        super().__init__(version, endpoint)
        self.data = data
        self.ignore_duplicates = ignore_duplicates

    def execute(self):
        sql = self.get_main_sql(self.data)
        ids = engine.execute(sql, len(self.data))
        ids = [x['id'] for x in ids]
        sql = self.get_m2m_sql(self.data, ids)
        engine.execute(sql)
        return ids

    def get_main_sql(self, data):
        assert isinstance(data, list)
        table = self.endpoint['table']
        fields = self.endpoint['fields']

        # Build the set of all field names used across all items to add.
        field_names = set()
        for single_data in data:
            for fn in single_data:
                if fn == 'id':
                    continue
                if fn not in fields:
                    raise ValueError(f'invalid field name: {fn}')
                fld = fields[fn]
                if fld.get('related', False) and 'm2m_table' in fld:
                    continue
                field_names.add(fn)

        # Construct the SQL.
        sql = f'INSERT INTO "{table}" (id'
        for fn in field_names:
            col = fields[fn]['column']
            sql += f',{col}'
        sql += f') VALUES '
        for ii, single_data in enumerate(data):
            if ii != 0:
                sql += ','
            id = single_data.get('id', 'DEFAULT')
            sql += f'({id}'
            for fn in field_names:
                fld = fields[fn]
                if fld.get('related', False):
                    if 'm2m_table' in fld:
                        continue
                if fn in single_data:
                    val = coerce(fld['field']['type'], single_data[fn])
                    val = engine.quote_literal(val)
                else:
                    val = 'DEFAULT'
                sql += f',{val}'
            sql += ')'

        # Ignore duplicates if requested.
        if self.ignore_duplicates:
            sql += ' ON CONFLICT DO NOTHING'

        # Return the IDs.
        sql += ' RETURNING id;'

        return sql

    def get_m2m_sql(self, data, ids):
        assert isinstance(data, list)
        fields = self.endpoint['fields']

        # Collect the set of m2m fields and the m2m table it goes with.
        m2ms = []
        for fn, fld in fields.items():
            fld = fields[fn]
            if fld.get('related', False) and 'm2m_table' in fld:
                m2ms.append((fn, fld['m2m_table']))

        # Contruct the sequence of inserts required to populate M2M
        # values.
        sql = ''
        for single_data, id in zip(data, ids):
            for fn, m2m_table in m2ms:
                val = single_data.get(fn, None)
                if not val:
                    continue
                if not isinstance(val, list):
                    raise ValueError('invalid many-to-many value, should be a list')
                sql += f'INSERT INTO "{m2m_table}" (lid, rid) VALUES '
                for ii, v in enumerate(val):
                    if not isinstance(v, int):
                        raise ValueError('invalid related ID, should be an integer')
                    if ii > 0:
                        sql += ','
                    sql += f'({id},{v})'
                sql += ';'

        return sql



from .query import Query
from .utils import strip_default


class List(Query):
    def __init__(self, model,
                 include=None, filter=None, order=None,
                 limit=None, include_limit=None,
                 offset=None, offset_reverse=False,
                 url=None,
                 **kwargs):
        super().__init__(model, **kwargs)
        self.include = include or []
        self.filter = filter
        self.order = order or ['id']  # default ordering for pagination
        self.limit = limit
        self.include_limit = include_limit
        self.offset = offset
        self.offset_reverse = offset_reverse
        self.url = url

    def iter_rows(self):
        for r in self.iter_main_rows():
            yield (
                'm',
                self._row_to_object(r)
            )
        for r in self.iter_included_rows():
            yield (
                'i',
                self._row_to_object(r)
            )
        if self.url:
            r = self.get_links_row()
            if r is not None:
                yield ('l', r)

    def iter_main_rows(self):
        raise NotImplementedError

    def iter_included_rows(self):
        raise NotImplementedError

    def iter_paths(self):
        if not hasattr(self, '_include_tree'):
            self._include_tree = self._make_include_tree()
        def _recurse(node, model, path=''):
            for k, v in node.items():
                try:
                    next_field = model.Meta.fields[k]
                except KeyError:
                    raise ValueError(
                        f'invalid include: {model} has no field {k}'
                    )
                try:
                    next_model = next_field.other
                except AttributeError:
                    raise ValueError(f'invalid include: included field is not a relationship')
                yield (k, path, next_model)
                for r in _recurse(v, next_model, f'{path}{k}.'):
                    yield r
        for r in _recurse(self._include_tree, self.model):
            yield r

    def _make_include_tree(self):
        tree = {}
        def _recurse(node, path, ii=0):
            p = path[ii]
            next = node.setdefault(p, {})
            if ii + 1 < len(path):
                _recurse(next, path, ii + 1)
        for i in self.include:
            path = i.split('.')
            _recurse(tree, path)
        return tree

    # async def respond(self, cursor):
    #     data = {'data': [], 'included': []}
    #     async for row in self.iter_main_rows(cursor):
    #         r = self._row_to_object(row)
    #         data['data'].append(r)
    #     unique = set()
    #     async for row in self.iter_included_rows(cursor):
    #         r = self._row_to_object(row, unique)
    #         if r:
    #             data['included'].append(r)
    #     if self.single:
    #         if len(data['data']):
    #             data['data'] = data['data'][0]
    #         else:
    #             data['data'] = None
    #     if not len(data['included']):
    #         del data['included']
    #     return data

    def _row_to_object(self, row, unique=None):
        id = row[1]['id']
        ident = (row[0], id)
        if unique:
            if ident in unique:
                return None
            unique.add(ident)
        ep = self.version[row[0]]
        attrs, rels = self._make_row_values(ep, row[1])
        # TODO: Need to decide on using system prefixed endpoints
        # or a flat namespace.
        obj = {
            'type': strip_default(row[0]),
            'id': id,
            'attributes': attrs,
            'relationships': rels
        }
        return obj

    def _make_row_values(self, model, row):
        attrs = {}
        rels = {}
        for field_name in model.Meta.fields.keys():
            value = row[field_name]
            field = model.Meta.fields[field_name]

            # The other type name should be that of the included endpoint.
            # However, there may not be an included endpoint, so what then?
            # Am currently setting to the value of the datapoint type, which
            # may be okay but seems suspect.
            # TODO: find alternative?
            other = getattr(field, 'other', None)

            if other:
                # TODO
                many = 'm2m_table' in field or field['field']['type'] == 'reverse'
                if many:
                    if value is None or value[0] is None:
                        rels[field_name] = {'data': []}
                    else:
                        rels[field_name] = {
                            'data': [
                                {'type': strip_default(other), 'id': v}
                                for v in value
                            ]
                        }
                else:
                    if value is None:
                        rels[field_name] = {'data': None}
                    else:
                        rels[field_name] = {
                            'data': {
                                'type': strip_default(other),
                                'id': value
                            }
                        }
            else:
                attrs[field_name] = value
        return attrs, rels



import shutil
from pathlib import Path

import click

from ..utils import get_data_dir
from .main import main

__all__ = ('helloworld',)


@main.group()
@click.pass_context
def example(ctx):
    pass


@example.command()
def helloworld():
    src = get_data_dir() / 'examples' / 'helloworld'
    dst = Path.cwd() / 'helloworld'
    shutil.copytree(src, dst)


@example.command()
def starwars():
    src = get_data_dir() / 'examples' / 'starwars'
    dst = Path.cwd() / 'starwars'
    shutil.copytree(src, dst)



from itertools import zip_longest

from .optiondict import Undefined


def chunkiter(iterable, size=10):
    args = [iter(iterable)]*size
    for ii in zip_longest(*args, fillvalue=Undefined):
        if tuple(ii)[-1] == Undefined:
            yield tuple(v for v in ii if v != Undefined)
        else:
            yield ii



from itertools import chain

from polecat.utils import to_tuple
from psycopg2.sql import SQL, Identifier

from .expression import Expression


class Select(Expression):
    def __init__(self, relation, columns=None, subqueries=None, joins=None,
                 where=None, limit=None, order=None):
        self.relation = relation
        self.columns = columns or ()
        self.subqueries = subqueries or {}
        self.joins = joins or ()
        self.where = where
        self.limit = limit
        self.order = to_tuple(order)

    @property
    def root_relation(self):
        return self.relation.root_relation

    def to_sql(self):
        columns_sql, columns_args = self.get_columns_sql()
        all_subquery_sql, all_subquery_args = self.get_subquery_sql()
        rel_sql, rel_args = self.relation.to_sql()
        joins_sql, joins_args = self.get_all_joins_sql()
        where_sql, where_args = self.get_where_sql()
        limit_sql = self.get_limit_sql()
        order_sql = self.get_order_sql()
        sql = SQL('SELECT {} FROM {}{}{}{}{}').format(
            SQL(', ').join(chain(columns_sql, all_subquery_sql)),
            rel_sql,
            joins_sql,
            where_sql,
            order_sql,
            limit_sql
        )
        return sql, columns_args + all_subquery_args + rel_args + joins_args + where_args

    def get_subquery_sql(self):
        all_subquery_sql = []
        all_subquery_args = ()
        for as_name, subquery in self.subqueries.items():
            subquery_sql, subquery_args = subquery.to_sql()
            sql = SQL('{} AS {}').format(
                subquery_sql,
                Identifier(as_name)
            )
            all_subquery_sql.append(sql)
            all_subquery_args += subquery_args
        return all_subquery_sql, all_subquery_args

    def get_columns_sql(self):
        columns_sql = []
        columns_args = ()
        for name in to_tuple(self.columns):
            if isinstance(name, Expression):
                sql, args = name.to_sql()
            elif isinstance(name, SQL):
                sql = name
                args = ()
            else:
                name_ident = Identifier(name)
                sql = SQL('{}.{} AS {}').format(
                    Identifier(self.relation.alias),
                    name_ident,
                    name_ident
                )
                args = ()
            columns_sql.append(sql)
            columns_args += args
        if not columns_sql and not self.subqueries:
            columns_sql = [SQL('*')]
        return columns_sql, columns_args

    def get_all_joins_sql(self):
        joins_sql = []
        joins_args = ()
        for join in self.joins:
            sql, args = self.get_join_sql(join)
            joins_sql.append(sql)
            joins_args += args
        if len(joins_sql):
            joins_sql = SQL(' ').join(joins_sql)
            return SQL(' {}').format(joins_sql), joins_args
        else:
            return SQL(''), ()

    def get_join_sql(self, join):
        return join.to_sql()

    def get_where_sql(self):
        if self.where:
            sql, args = self.where.get_sql(self.relation)
            return SQL(' WHERE {}').format(sql), args
        else:
            return SQL(''), ()

    def get_limit_sql(self):
        if self.limit is not None:
            # TODO: Is this risky?
            return SQL(f' LIMIT {int(self.limit)}')
        else:
            return SQL('')

    def get_order_sql(self):
        if self.order:
            columns = ', '.join(self.parse_order_column(o) for o in self.order)
            return SQL(f' ORDER BY {columns}')
        else:
            return SQL('')

    def parse_order_column(self, column):
        if column[0] == '-':
            return f'"{column[1:]}" DESC'
        else:
            return f'"{column}"'

    def iter_column_names(self):
        for name in self.columns:
            yield name
        for name in self.subqueries.keys():
            yield name

    def get_column(self, name):
        # TODO: Not sure a select should be running "get_column".
        return self.relation.get_column(name)

    def get_subrelation(self, name):
        return self.relation.get_subrelation(name)

    def push_selection(self, selection=None):
        from ...schema import ReverseColumn
        # TODO: Efficiency. Everything.
        for column_name in selection or ():
            if column_name not in self.columns and column_name not in self.subqueries:
                self.columns += (column_name,)
        # TODO: Forgot that columns can be a subquery. I'll need to
        # allow subqueries to return a list columns they require. For
        # now, skip it.
        to_push = self.columns if isinstance(self.columns, (tuple, list)) else ()
        for column_name in self.subqueries.keys():
            try:
                column = self.relation.get_column(column_name)
            except KeyError:
                continue
            if not isinstance(column, ReverseColumn):
                to_push += (column_name,)
        if self.where:
            to_push += self.where.get_primary_columns()
        self.relation.push_selection(to_push)
        for join in self.joins:
            join.push_selection()



import inspect
import random
import string
from pathlib import Path

from .retry import retry  # noqa
from .stringcase import snakecase


class Undefined:
    pass


def add_attribute(obj, key, value):
    setattr(obj, key, value)
    return obj


def random_ident(length=8):
    return ''.join(
        random.choices(string.ascii_lowercase, k=1) +
        random.choices(
            string.ascii_lowercase + string.digits,
            k=length
        )
    )


def to_class(value):
    return value if inspect.isclass(value) else value.__class__


def to_bool(value):
    if isinstance(value, str):
        return not (
            not len(value) or value.lower()[0] in ('f', 'n', '0')
        )
    else:
        return bool(value)


def get_class_path(value):
    return Path(inspect.getfile(value)).parent


def to_tuple(value, keep_none=False):
    if isinstance(value, tuple):
        return value
    if isinstance(value, list):
        return tuple(value)
    elif value is None:
        if keep_none:
            return (None,)
        else:
            return ()
    else:
        return (value,)


def to_list(value):
    # TODO: Meh, could be faster.
    return list(to_tuple(value))


def to_set(value):
    if isinstance(value, set):
        return value
    elif value is None:
        return set()
    else:
        return set(value)


def to_dict(value):
    if isinstance(value, None):
        return {}
    elif not isinstance(value, dict):
        return dict(value)
    else:
        return value


def merge(target, source):
    for k, v in source.items():
        if isinstance(v, dict):
            merge(target.setdefault(k, {}), v)
        else:
            target[k] = v
    return target


def substitute(template, **kwargs):
    return string.Template(template).safe_substitute(kwargs)


def capitalize(value):
    if value:
        return value[0].upper() + value[1:]
    else:
        return value


def uncapitalize(value):
    if value:
        return value[0].lower() + value[1:]
    else:
        return value


def get_path(source, path, separator='.'):
    parts = path.split(separator)
    value = source
    for part in parts:
        if value is not None:
            value = value.get(part)
    return value


def set_path(destination, path, value, separator='.'):
    parts = path.split(separator)
    for part in parts[:-1]:
        if part not in destination:
            destination[part] = {}
        destination = destination[part]
    destination[parts[-1]] = value


def get_data_dir():
    return Path(__file__).parent.parent / 'data'


def indent(text, size=4):
    return ('\n' + ' ' * size).join(text.split('\n'))


def name_from_class(cls, suffix=None):
    name = snakecase(to_class(cls).__name__)
    if suffix:
        pos = -len(suffix) - 1
        if name[pos:] == '_' + suffix:
            name = name[:pos]
    return name



import click

from .feedback import HaloFeedback, cli_feedback
from .main import main

__all__ = ('build',)


@main.command()
@click.option('--source')
@click.option('--local-package', '-l', multiple=True)
@click.pass_context
@cli_feedback('Build server code')
def build(ctx, source, local_package, feedback=None):
    from ..deploy.aws.build import build as aws_build
    project = ctx.obj['project']
    aws_build(project, source, local_package, feedback=feedback)



from psycopg2.sql import SQL, Placeholder

from .expression.exists import Exists
from .expression.expression import Expression
from .expression.insert import Insert
from .expression.not_ import Not
from .expression.select import Select
from .expression.subquery import Subquery
from .expression.union import Union
from .expression.where import Where
from .insert_strategy import InsertStrategy


# TODO: This class should be merged into Select.
class DummySelect(Expression):
    def __init__(self, values, where):
        self.values = values
        self.where = where

    def to_sql(self):
        where_sql, where_args = self.where.to_sql()
        # TODO: This is all wrong, need to think more clearly about
        # this.
        values_sql = []
        values_args = ()
        for value in self.values.values():
            if isinstance(value, Expression):
                sql, args = value.to_sql()
                values_sql.append(sql)
                values_args += args
            else:
                values_sql.append(Placeholder())
                values_args += (value,)
        values_sql = SQL(', ').join(values_sql)
        return (
            SQL('SELECT {} WHERE {}').format(
                values_sql,
                where_sql
            ),
            values_args + where_args
        )

    def iter_column_names(self):
        return self.values.keys()


class InsertIfMissingStrategy(InsertStrategy):
    def __init__(self, root_strategy):
        self.root = root_strategy
        self.relation_counter = 0

    def parse_query(self, query):
        parsed_values = self.parse_values(query, query.values)
        # TODO: This is pretty awful.
        parsed_values = parsed_values.values.as_dict()
        values = {
            **query.defaults,
            **parsed_values
        }
        where_expr = Where(**parsed_values)
        insert_expr = Insert(
            query.source,
            DummySelect(
                values,
                where=Not(
                    Exists(
                        Subquery(
                            Select(
                                query.source,
                                (SQL('1'),),
                                where=where_expr
                            )
                        )
                    )
                )
            )
        )
        alias = self.root.create_alias(insert_expr)
        expr = Union([
            Select(alias),
            Select(
                query.source,
                ('id',),  # TODO: Wat...
                where=where_expr
            )
        ])
        return expr



from base64 import b64decode

import ujson

from ...utils import capitalize
from ...utils.feedback import feedback
from .utils import aws_client


@feedback
def run_command(project, deployment, command, args, kwargs, feedback):
    lmd = aws_client('lambda')
    project_deployment = f'{capitalize(project)}{capitalize(deployment)}'
    payload = ujson.dumps({
        'event': 'admin',
        'command': command,
        'args': args,
        'kwargs': kwargs
    }).encode()
    response = lmd.invoke(
        FunctionName=f'{project_deployment}Server',
        LogType='Tail',
        Payload=payload
    )
    logs = response.get('LogResult')
    if logs:
        # TODO: Move to somewhere better.
        logs = b64decode(logs)
        logs = logs.replace(b'\r\xc2\xa0\xc2\xa0', b'\n')
        print(logs.decode())



class Handler:
    def __init__(self, project, middleware=None):
        self.project = project
        self.middleware = middleware or []

    def prepare(self):
        pass

    def match(self, event):
        return False

    async def run(self, event):
        for mw in self.middleware:
            mw.run(event)
        return await self.handle_event(event)

    async def handle_event(self, event):
        raise NotImplementedError


class HTTPHandler(Handler):
    def match(self, event):
        return event.is_http()



from polecat.utils import to_tuple
from psycopg2.sql import SQL, Identifier

from ..query.selection import Selection
from ..schema import ReverseColumn
from .expression.alias import Alias
from .expression.array_agg import ArrayAgg
from .expression.as_ import As
from .expression.join import Join, LateralJoin
from .expression.raw import RawSQL
from .expression.select import Select
from .expression.subquery import Subquery
from .expression.where import Where


class SelectStrategy:
    def __init__(self, root_strategy):
        self.root = root_strategy
        self.relation_counter = 0
        self.lateral_counter = 0
        self.agg_counter = 0

    def parse_query(self, queryable):
        expr = self.parse_query_from_components(
            queryable.source,
            queryable.selection,
            queryable=queryable
        )
        if queryable.recurse_column:
            alias = self.root.create_alias(expr)
            top_level = Select(
                alias,
                columns=expr.columns + tuple(expr.subqueries.keys())
            )
            alias.push_selection(Selection(queryable.recurse_column))
            cte_as = alias.expression
            join = Join(alias)
            recursive_expr = self.parse_query_from_components(
                queryable.source,
                queryable.selection,
                queryable=queryable
            )
            recursive_expr.push_selection(Selection(queryable.recurse_column))
            # TODO: This is the worst.
            if isinstance(recursive_expr.relation, As):
                expr_to_update = recursive_expr.relation.expression.expression
            else:
                expr_to_update = recursive_expr
            expr_to_update.joins = to_tuple(expr_to_update.joins) + (join,)
            expr_to_update.where = Where(
                id=SQL('{}.{}').format(
                    Identifier(alias.alias),
                    Identifier(queryable.recurse_column)
                )
            )
            cte_as.set_recursive_expression(recursive_expr)
            expr = top_level
        return expr

    def parse_query_from_components(self, relation, selection, queryable=None):
        relation = self.parse_relation(relation)
        if selection.has_lookups():
            relation = self.create_alias_for_relation(relation)
        columns = selection.fields
        subqueries, joins = self.create_subqueries(relation, selection)
        return Select(
            relation, columns, subqueries, joins,
            limit=queryable.limit if queryable else None,
            order=queryable.order if queryable else None
        )

    def parse_relation(self, relation):
        return self.root.parse_chained_relation(relation)

    def create_alias_for_relation(self, relation):
        # TODO: Using the factory here when we've already got an "As"
        # spuriously increments the relation counter.
        alias = As.factory(relation, f't{self.relation_counter}')
        self.relation_counter += 1
        return alias

    def create_subqueries(self, relation, selection):
        lookups = selection.lookups
        subqueries = {}
        joins = []
        for name, subquery in lookups.items():
            if isinstance(subquery, Selection):
                expr = self.create_lateral_subquery(relation, subquery, name)
                joins.append(expr)
                # TODO: This is ugly as. The problem is that
                # subqueries aren't necessarily columns, they can be
                # just named subqueries. Need to separate them.
                try:
                    column = relation.get_column(name)
                    if isinstance(column, ReverseColumn):
                        alias_column = 'array_agg'
                    else:
                        alias_column = None
                except KeyError:
                    alias_column = None
                subqueries[name] = Alias(expr.alias, column=alias_column)
            else:
                expr = self.create_detached_subquery(subquery)
                subqueries[name] = expr
        return subqueries, joins

    def create_lateral_subquery(self, relation, selection, column_name):
        subrelation = relation.get_subrelation(column_name)
        subquery = self.parse_query_from_components(subrelation, selection)
        self.add_where_clause_to_lateral_subquery(relation, column_name, subquery)
        subquery = self.add_aggregation_to_lateral_subquery(relation, column_name, subquery)
        lateral_alias = self.create_alias_name_for_lateral()
        return LateralJoin(
            subquery,
            join_type=LateralJoin.LEFT,
            alias=lateral_alias,
            condition=self.create_lateral_condition(
                relation,
                lateral_alias,
                column_name
            )
        )

    def add_where_clause_to_lateral_subquery(self, relation, column_name, subquery):
        column = relation.get_column(column_name)
        if isinstance(column, ReverseColumn):
            subquery.where = Where(**{
                column.related_column.name: SQL('{}.id').format(
                    Identifier(relation.alias)
                )
            })
        else:
            subquery.where = Where(**{
                'id': SQL('{}.{}').format(
                    Identifier(relation.alias),
                    Identifier(column_name)
                )
            })

    def add_aggregation_to_lateral_subquery(self, relation, column_name, subquery):
        column = relation.get_column(column_name)
        if isinstance(column, ReverseColumn):
            agg_alias = self.create_alias_name_for_aggregate()
            subquery = Select(
                As(Subquery(subquery), agg_alias),
                columns=ArrayAgg(Alias(agg_alias))
            )
        return subquery

    def create_lateral_condition(self, relation, lateral_alias, column_name):
        # TODO: Replace this with something other than a raw sql string.
        column = relation.get_column(column_name)
        if isinstance(column, ReverseColumn):
            return RawSQL(SQL('TRUE'))
        else:
            return RawSQL(SQL('TRUE'))

    def create_detached_subquery(self, subquery):
        return Subquery(self.root.parse_queryable_or_builder(subquery))

    def create_alias_name_for_lateral(self):
        alias_name = f'j{self.lateral_counter}'
        self.lateral_counter += 1
        return alias_name

    def create_alias_name_for_aggregate(self):
        alias_name = f'a{self.agg_counter}'
        self.agg_counter += 1
        return alias_name



from psycopg2.sql import SQL, Identifier

from .expression import Expression
from .insert import Insert


class Update(Insert):
    def __init__(self, *args, where=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.where = where

    @property
    def alias(self):
        return self.relation.alias

    def to_sql(self):
        if isinstance(self.values, Expression):
            prefix_sql = SQL('')
            suffix_sql = SQL('')
            # TODO: This is ugly as.
            self.values.keyword = 'ROW'
            get_values_func = self.get_values_sql_from_expression
        else:
            prefix_sql = SQL('ROW (')
            suffix_sql = SQL(')')
            get_values_func = self.get_values_sql_from_dict
        column_names_sql, column_values_sql, column_values = get_values_func()
        returning_sql = map(Identifier, self.returning)
        where_sql, where_args = self.get_where_sql()
        sql = SQL('UPDATE {} SET ({}) = {}{}{}{} RETURNING {}').format(
            Identifier(self.relation.alias),
            SQL(', ').join(column_names_sql),
            prefix_sql,
            SQL(', ').join(column_values_sql),
            suffix_sql,
            where_sql,
            SQL(', ').join(returning_sql)
        )
        return sql, tuple(column_values) + where_args

    def get_where_sql(self):
        if self.where:
            sql, args = self.where.get_sql(self.relation)
            return SQL(' WHERE {}').format(sql), args
        else:
            return SQL(''), ()



from .filter import Filter
from .multi.list import List


class Detail(List):
    def __init__(self, version, endpoint, id, include=[], where=[], **kwargs):
        self.id = id
        table = endpoint['table']
        flt = Filter({'id': id})
        super().__init__(version, endpoint, include, filter=flt, limit=1, **kwargs)



import click
from termcolor import colored

from .config import update_config
from .feedback import HaloFeedback
from .main import main

__all__ = ('create_project', 'list_projects')


@main.group()
@click.pass_context
def project(ctx):
    pass


@project.command('create')
@click.argument('project')
@click.pass_context
def create_project(ctx, project):
    from ..deploy.aws.project import create_project as aws_create_project
    bucket = ctx.obj['bucket']
    aws_create_project(project, bucket, feedback=HaloFeedback())
    update_config({'project': project})


@project.command('list')
def list_projects():
    from ..deploy.aws.project import list_projects as aws_list_projects
    projects = aws_list_projects(feedback=HaloFeedback())
    if len(projects) == 0:
        print(colored('    No projects', 'grey'))
    else:
        for proj in projects:
            print(f'    {proj}')


@project.command('delete')
@click.argument('project')
@click.pass_context
def delete_project(ctx, project):
    from ..deploy.aws.project import delete_project as aws_delete_project
    aws_delete_project(project, feedback=HaloFeedback())
    update_config(delete=['project'])



from polecat.db.query import Q as BaseQ
from polecat.db.query import S

from ..model import Model
from .helpers import model_to_values, set_values_on_model

__all__ = ('Q', 'S')


class Q(BaseQ):
    def __init__(self, model=None, queryable=None, branches=None, session=None):
        super().__init__(
            queryable or (model and model.Meta.table) or None,
            branches,
            session
        )
        self.model = model

    def insert(self, model_or_subquery=None, **kwargs):
        if isinstance(model_or_subquery, Model):
            if kwargs:
                raise ValueError('Cannot pass both model and values to insert')
            return self.insert_from_model_argument(model_or_subquery)
        elif not kwargs and isinstance(self.model, Model):
            # TODO: Ensure test for this.
            # if not isinstance(self.model, Model):
            #     raise ValueError(
            #         'Cannot pass nothing to insert when root queryable is not'
            #         ' a Model'
            #     )
            return self.insert_from_root()
        else:
            args = (model_or_subquery,) if (model_or_subquery is not None) else ()
            return super().insert(*args, **kwargs)

    def insert_if_missing(self, values_or_model=None, defaults={}):
        if isinstance(values_or_model, Model):
            return super().insert_if_missing(
                model_to_values(self.model),
                defaults=defaults
            )
        elif values_or_model is not None:
            return super().insert_if_missing(
                values_or_model,
                defaults=defaults
            )
        else:
            return super().insert_if_missing(
                model_to_values(self.model),
                defaults=defaults
            )

    def insert_from_model_argument(self, model):
        return super().insert(**model_to_values(model))

    def insert_from_root(self):
        return super().insert(**model_to_values(self.model))

    def update(self, model_or_subquery=None, **kwargs):
        if isinstance(model_or_subquery, Model):
            if kwargs:
                raise ValueError('Cannot pass both model and values to update')
            return self.update_from_model_argument(model_or_subquery)
        elif not kwargs:
            if not isinstance(self.model, Model):
                raise ValueError(
                    'Cannot pass nothing to update when root queryable is not'
                    ' a Model'
                )
            return self.update_from_root()
        else:
            return super().update(**kwargs)

    def update_from_model_argument(self, model):
        id = self.get_model_id(model)
        return (
            super()
            .filter(id=id)
            .update(**model_to_values(model, exclude_fields=('id',)))
        )

    def update_from_root(self):
        id = self.get_model_id(self.model)
        return (
            super()
            .filter(id=id)
            .update(**model_to_values(self.model, exclude_fields=('id',)))
        )

    def get_model_id(self, model):
        id = getattr(model, 'id', None)
        if id is None:
            raise ValueError('Model has no ID set.')
        return id

    def into(self, destination_model):
        for row in self:
            set_values_on_model(row, destination_model)
            return destination_model
        raise ValueError('No results returned from query')

    def chain(self, queryable):
        return self.__class__(self.model, queryable, self.branches, self.session)



import inspect
from operator import attrgetter

from polecat.utils import to_class
from polecat.utils.container import Option, OptionDict, passthrough

from .context import active_context

__all__ = ('RegistryMetaclass', 'Registry', 'MappedRegistry')

active_context().Meta.add_options(
    Option('registries', default=passthrough(OptionDict)())
)


class RegistryMetaclass(type):
    @classmethod
    def is_baseclass(cls, name, bases):
        return not (
            bases and
            name != getattr(bases[0], '_registry_base', None)
        )

    def __init__(cls, name, bases, dct):
        if not cls.is_baseclass(name, bases):
            active_context().registries[cls._registry].add(cls)
        super().__init__(name, bases, dct)


class Registry:
    def __init__(self, name, construct=False, name_mapper=None):
        self.name = name
        self.values = []
        self.construct = construct
        self.name_mapper = name_mapper or attrgetter('name')
        self.name_map = {}
        active_context().registries.Meta.add_options(
            Option(self.name, default=self)
        )

    def __iter__(self):
        return self.values.__iter__()

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.values[key]
        else:
            return self.name_map[key]

    def add(self, value):
        if self.construct:
            value = value()
        self.values.append(value)
        if self.name_mapper:
            self.name_map[self.name_mapper(value)] = value


class MappedRegistry(Registry):
    def __init__(self, name):
        super().__init__(name)
        self.values = {}

    def __getitem__(self, value):
        for base_class in inspect.getmro(to_class(value)):
            candidates = self.values.get(base_class)
            if candidates:
                for c in candidates:
                    match = getattr(c, 'match', None)
                    if not match or match(value):
                        return c
        raise KeyError(f'No {self.name} mapping for {value}')

    def add(self, mapped_value):
        for src in mapped_value.sources:
            self.values.setdefault(to_class(src), []).append(mapped_value)



import importlib


def import_attribute(path_or_attr):
    if isinstance(path_or_attr, str):
        last_index = path_or_attr.rfind('.')
        module_path = path_or_attr[:last_index]
        attr_name = path_or_attr[last_index + 1:]
        module = importlib.import_module(module_path)
        path_or_attr = getattr(module, attr_name)
    return path_or_attr



from psycopg2.sql import SQL, Identifier

from .expression import Expression


class StringAgg(Expression):
    def __init__(self, expression, separator=' '):
        self.expression = expression
        self.separator = separator

    @property
    def alias(self):
        return self.expression.alias

    def to_sql(self):
        expr_sql, expr_args = self.expression.to_sql()
        return (
            SQL('string_agg({}, {})').format(
                expr_sql, Identifier(self.separator)
            ),
            expr_args
        )

    def get_subrelation(self, name):
        # TODO: This feels like bad design. Not sure how to correct
        # just at the moment, and at least it's clear what it's doing.
        return self.expression.get_subrelation(name)

    def get_column(self, name):
        # TODO: Also bad design, but also what's the difference
        # between this and the above? Looks like the above just does
        # 'this.related_table'.
        return self.expression.get_column(name)

    def has_column(self, name):
        # TODO: Also bad design, but also what's the difference
        # between this and the above? Looks like the above just does
        # 'this.related_table'.
        return self.expression.has_column(name)

    def push_selection(self, selection=None):
        self.expression.push_selection(selection)



from polecat.utils.repr import to_repr

from ..role_prefix import get_role_prefix
from .entity import Entity


class Role(Entity):
    def __init__(self, name, parents=None, options=None, app=None):
        self.app = app
        self.name = name
        self.parents = parents or ()
        self.options = options or {}
        self.schema = None

    def __repr__(self):
        return to_repr(
            self,
            name=self.name
        )

    @property
    def dbname(self):
        prefix = get_role_prefix()
        if prefix:
            return f'{prefix}_{self.name}'
        else:
            return self.name

    @property
    def signature(self):
        return (Role, self.name)

    def has_changed(self, other):
        if isinstance(other, str):
            return self.name != other
        else:
            return (
                self.name != other.name or
                self.has_changed_parents(other)
            )

    def has_changed_parents(self, other):
        # TODO: I don't like this much.
        return (
            sorted(getattr(p, 'name', p) for p in self.parents) !=
            sorted(getattr(p, 'name', p) for p in other.parents)
        )

    def bind(self, schema):
        if self.schema is None:
            self.schema = schema
            self.parents = [
                p if isinstance(p, Role) else schema.get_role_by_name(p)
                for p in self.parents
            ]


class Access(Entity):
    def __init__(self, entity, all=None, select=None, insert=None, update=None,
                 delete=None, app=None):
        self.entity = entity
        self.all = all or ()
        self.select = select or ()
        self.insert = insert or ()
        self.update = update or ()
        self.delete = delete or ()
        self.app = app
        self.schema = None

    def __repr__(self):
        return to_repr(
            self,
            name=self.entity if isinstance(self.entity, str) else self.entity.name  # TODO: Need an interface.
        )

    @property
    def signature(self):
        return (Access, self.entity.signature)

    def has_changed(self, other):
        return (
            self.entity != other.entity or
            self.all != other.all or
            self.select != other.select or
            self.insert != other.insert or
            self.update != other.update or
            self.delete != other.delete
        )

    def bind(self, schema):
        if self.schema is None:
            self.schema = schema
            # TODO: Cleanup
            self.all = [
                p if isinstance(p, Role) else schema.get_role_by_name(p)
                for p in self.all
            ]
            self.select = [
                p if isinstance(p, Role) else schema.get_role_by_name(p)
                for p in self.select
            ]
            self.insert = [
                p if isinstance(p, Role) else schema.get_role_by_name(p)
                for p in self.insert
            ]
            self.update = [
                p if isinstance(p, Role) else schema.get_role_by_name(p)
                for p in self.update
            ]
            self.delete = [
                p if isinstance(p, Role) else schema.get_role_by_name(p)
                for p in self.delete
            ]



from .model import Mutation


def create_mutation(name, resolver, returns=None, input=None):
    cls = type(name, (Mutation,), {
        'returns': returns,
        'input': input
    })
    cls.resolve = classmethod(resolver)
    return cls



from psycopg2.sql import SQL, Placeholder

from .expression import Expression


def value_to_sql(column, value):
    if isinstance(value, Expression):
        sql, args = value.to_sql()
        return SQL('{}').format(sql), args
    else:
        return Placeholder(), (column.to_db_value(value),)



from psycopg2.sql import SQL, Identifier


class Join:
    LEFT = 'LEFT'

    def __init__(self, expression, join_type=None, alias=None, condition=None):
        self.expression = expression
        self.join_type = join_type
        self.alias = alias
        self.condition = condition

    def to_sql(self):
        expr_sql, expr_args = self.expression.to_sql()
        if self.join_type is None:
            return SQL(', {}').format(expr_sql), expr_args
        join_type = self.get_join_type_fragment()
        cond_sql, cond_args = self.get_condition_sql()
        alias_sql = self.get_alias_sql()
        return (
            SQL('%sJOIN {}{}{}' % join_type).format(
                expr_sql,
                alias_sql,
                cond_sql
            ),
            expr_args + cond_args
        )

    def get_join_type_fragment(self):
        join_type = self.join_type
        if join_type:
            join_type += ' '
        return join_type

    def get_alias_sql(self):
        if self.alias:
            return SQL(' AS {}').format(Identifier(self.alias))
        else:
            return SQL('')

    def get_condition_sql(self):
        if self.condition:
            cond_sql, cond_args = self.condition.to_sql()
            return (
                SQL(' ON {}').format(cond_sql),
                cond_args
            )
        else:
            return SQL(''), ()

    def push_selection(self, selection=None):
        self.expression.push_selection(selection=selection)


class LateralJoin(Join):
    def to_sql(self):
        expr_sql, expr_args = self.expression.to_sql()
        join_type = self.get_join_type_fragment()
        cond_sql, cond_args = self.get_condition_sql()
        return (
            SQL('%sJOIN LATERAL ({}) AS {}{}' % join_type).format(
                expr_sql,
                Identifier(self.alias),
                cond_sql
            ),
            expr_args + cond_args
        )



import os
from pathlib import Path

from termcolor import colored

from ...utils.feedback import feedback
from .exceptions import KnownError
from .operations import get_parameter, set_parameter
from .utils import aws_client


@feedback
def upload(project, bucket, source=None, version=None, feedback=None):
    s3 = aws_client('s3')
    source = Path(source or os.getcwd())
    version_path = f'/polecat/projects/{project}/code/version'  # TODO: Constant.
    current_version = None
    with feedback(f'Prepare version number') as fb:
        current_version = get_parameter(version_path)
        if current_version is not None:
            current_version = int(current_version)
        else:
            current_version = 0
        if version is not None:
            version = int(version)
            if version > current_version + 1:
                raise KnownError('invalid version provided')
        else:
            version = current_version + 1
        fb.message = f'Using version {colored(version, "blue")}'
    for file in ('server.zip',):  # 'admin.zip'):
        with feedback(f'Upload {colored(file, "blue")}'):
            s3.upload_file(
                str(source / file),
                Bucket=bucket,
                Key=f'projects/{project}/code/{version}/{file}'  # TODO: Constant.
            )
    if version > current_version:
        with feedback(f'Set current version to {colored(version, "blue")}'):
            set_parameter(version_path, str(version))


@feedback
def upload_bundle(project, bucket, source=None, version=None, feedback=None):
    s3 = aws_client('s3')
    source = Path(source or (Path(os.getcwd()) / 'dist'))
    def _upload(path, dst_file):
        params = {}
        if '.map.' in dst_file or dst_file.endswith('.map'):
            return
        # TODO: Should really generalize this.
        if dst_file.endswith('stats.json'):
            return
        if dst_file.endswith('.br'):
            return
        if dst_file.endswith('.gz'):
            params['ExtraArgs'] = {
                'ContentEncoding': 'gzip'
            }
        with feedback(f'Upload {colored(dst_file, "blue")}'):
            s3.upload_file(
                path,
                Bucket=bucket,
                Key=f'projects/{project}/bundle/{version}/{dst_file}',  # TODO: Constant.
                **params
            )
        uploaded.append(dst_file)
    # TODO: This is duplication of above.
    version_path = f'/polecat/projects/{project}/bundle/version'  # TODO: Constant.
    current_version = None
    with feedback(f'Prepare version number') as fb:
        current_version = get_parameter(version_path)
        if current_version is not None:
            current_version = int(current_version)
        else:
            current_version = 0
        if version is not None:
            version = int(version)
            if version > current_version + 1:
                raise KnownError('invalid version provided')
        else:
            version = current_version + 1
        fb.message = f'Using version {colored(version, "blue")}'
    uploaded = []
    if source.is_file():
        _upload(str(source), source.name)
    for root, dirs, files in os.walk(source):
        for file in files:
            # TODO: Filter out certain files?
            path = str(Path(root, file))
            dst_file = path[len(str(source)):]
            if dst_file[0] == '/':
                dst_file = dst_file[1:]
            _upload(path, dst_file)
    # TODO: Duplication of above.
    if version > current_version:
        with feedback(f'Set current version to {colored(version, "blue")}'):
            if not len(uploaded):
                raise KnownError('no files uploaded')
            set_parameter(version_path, str(version))



import os
from contextlib import contextmanager
from urllib.parse import urlparse

import psycopg2


class ConnectionManager:
    def __init__(self):
        # TODO: This should be transferred to POLECAT_DATABASE_URL
        # from the config.
        self.stack = [
            x
            for x in [os.environ.get('DATABASE_URL')]
            if x is not None
        ]
        self.connections = {}
        self.current_cursor = {}

    def get_url(self, url=None):
        try:
            return url or self.stack[-1]
        except IndexError:
            # TODO: Better exception.
            raise Exception('No database url specified')

    # TODO: autocommit really doesn't make sense here for the
    # manager. It needs to be in the __init__ method, as the
    # connection gets retained across calls.
    @contextmanager
    def connection(self, url=None, autocommit=True):
        url = self.get_url(url)
        conn = self.connections.get(url)
        if not conn:
            conn = self.open_connection(url, autocommit=autocommit)
        try:
            yield conn
            if not autocommit:
                conn.commit()
        except Exception:
            # TODO: Do I really need this?
            conn.rollback()
            raise

    @contextmanager
    def cursor(self, url=None, cursor=None, autocommit=True):
        # TODO: This conditional is ugly.
        if cursor:
            yield cursor
        url = self.get_url(url)
        curs = self.current_cursor.get(url)
        if curs:
            yield curs
        else:
            with self.connection(url, autocommit=autocommit) as conn:
                curs = conn.cursor()
                self.current_cursor[url] = curs
                try:
                    yield curs
                finally:
                    curs.close()
                    del self.current_cursor[url]

    def open_connection(self, url, autocommit=True):
        url = self.get_url(url)
        conn = psycopg2.connect(url)
        if autocommit:
            conn.autocommit = True
        self.connections[url] = conn
        return conn

    def close_all_connections(self):
        all_urls = list(self.connections.keys())
        for url in all_urls:
            self.close_connection(url)

    def close_connection(self, url):
        try:
            self.connections[url].close()
        except KeyError:
            pass
        try:
            del self.connections[url]
        except KeyError:
            pass

    @contextmanager
    def push_url(self, url):
        self.stack.append(url)
        try:
            yield
        finally:
            self.pop_url()

    def pop_url(self):
        url = self.stack.pop()
        self.close_connection(url)

    def parse_url(self, url=None):
        url = self.get_url(url)
        info = urlparse(url)
        return {
            'scheme': info.scheme,
            'dbname': info.path[1:],
            'host': info.hostname,
            'port': info.port,
            'username': info.username,
            'password': info.password
        }


manager = ConnectionManager()
connection = manager.connection
cursor = manager.cursor


@contextmanager
def transaction(curs=None):
    # TODO: Not super happy about how manual the BEGIN and COMMIT
    # calls are. Performance hit?
    if curs is None:
        with cursor() as curs:
            try:
                curs.execute('BEGIN')
                yield curs
                curs.execute('COMMIT')
            except Exception:
                curs.execute('ROLLBACK')
                raise
    else:
        try:
            curs.execute('BEGIN')
            yield curs
            curs.execute('COMMIT')
        except Exception:
            curs.execute('ROLLBACK')
            raise



from .expression import Expression


class RawSQL(Expression):
    def __init__(self, sql, args=None):
        self.sql = sql
        self.args = args or ()

    def to_sql(self):
        return self.sql, self.args



from psycopg2.sql import SQL, Identifier


class Index:
    def __init__(self, column_name):
        self.column_name = column_name

    def bind(self, table):
        self.table = table
        self.name = '{}_{}_{}'.format(
            self.table.name,
            self.column_name,
            self._get_suffix()
        )

    @property
    def sql(self):
        return SQL('CREATE INDEX {} ON {} ({})').format(
            Identifier(self.name),
            Identifier(self.table.name),
            Identifier(self.column_name)
        ), ()

    def serialize(self):
        return 'index.Index(column_name={})'.format(
            repr(self.column_name)
        )

    def _get_suffix(self):
        return 'idx'


class GistTrigramIndex(Index):
    @property
    def sql(self):
        return SQL('CREATE INDEX {} ON {} USING GIST ({} gist_trgm_ops)').format(
            Identifier(self.name),
            Identifier(self.table.name),
            Identifier(self.column_name)
        ), ()

    def serialize(self):
        return 'index.GistTrigramIndex(column_name={})'.format(
            repr(self.column_name)
        )

    def _get_suffix(self):
        return 'trgm_idx'



from polecat.utils.config import Config
from polecat.utils.optiondict import Option
from polecat.utils.proxy import Proxy


class RootConfig(Config):
    debug = (bool, False)
    log_sql = (bool, False)
    jwt_secret = str
    database_url = str


def mount_config(config, path):
    segments = path.split('.')
    cur_config = default_config
    for segment in segments[:-1]:
        cur_config = getattr(cur_config, segment)
    if cur_config.Meta.has_option(segments[-1]):
        cur_config = getattr(cur_config, segments[-1])
        for name, option in config.Meta.options.items():
            cur_config.Meta.add_option(option)
    else:
        prefix = ''.join([cur_config.Meta.prefix, segments[-1].upper()])
        config.Meta.set_prefix(prefix)
        cur_config.Meta.add_option(Option(segments[-1], config))


def auto_mount_config(path):
    def inner(cls):
        mount_config(cls(), path)
        return cls
    return inner


default_config = Proxy(lambda: RootConfig(prefix='POLECAT'))



from .insert import Insert  # noqa
from .multi import Multi  # noqa
from .select import Select  # noqa
from .subquery import Subquery  # noqa
from .update import Update  # noqa
from .values import Values  # noqa
from .variable import LocalRole, LocalVariable  # noqa
from .where import Where  # noqa



from ..core.context import active_context
from ..project.handler import Handler


class AdminHandler(Handler):
    def match(self, event):
        return event.is_admin()

    async def handle_event(self, event):
        command = active_context().registries.command_registry[event.event['command']]
        args = event.event['args']
        kwargs = event.event['kwargs']
        command(self.project).run(*args, **kwargs)



from ..query import query as query_module
from .expression.update import Update
from .expression.where import Where
from .expression.subrelation_override import SubrelationOverride
from .insert_strategy import InsertStrategy


class UpdateStrategy(InsertStrategy):
    def parse_query(self, query):
        # TODO: Don't like this.
        if isinstance(query.source, query_module.Filter):
            relation = query.source.source
            where = Where(**query.source.options)
        else:
            relation = query.source
            where = None
        expr = Update(
            relation,
            self.parse_values_or_subquery(query, query.values),
            where=where
        )
        if query.reverse_queries:
            expr = SubrelationOverride(expr, query.reverse_queries, self.root)
        return expr



from psycopg2.sql import SQL


class Not:
    def __init__(self, expression):
        self.expression = expression

    def to_sql(self):
        expr_sql, expr_args = self.expression.to_sql()
        return SQL('NOT {}').format(expr_sql), expr_args



from stdb.core.auth import check_access
from stdb.core.cache import default_cache as cache
from stdb.exceptions import StdbError
from stdb.utils import split_handle

from .bulk_create import BulkCreate
from .bulk_update import BulkUpdate
from .create import Create
from .delete import Delete
from .detail import Detail
from .filter import Filter
from .multi.list import List
from .update import Update


class DoesNotExist(Exception):
    pass


def create(handle, data, id=None):
    ver, ep = cache.get_version_and_endpoint(handle)
    check_access(ep, 'create')
    if id is None:
        qry = Create(ver, ep, data)
    else:
        qry = Update(ver, ep, id, data)
    return qry.execute()


def update(handle, id, data):
    ver, ep = cache.get_version_and_endpoint(handle)
    check_access(ep, 'update')
    qry = Update(ver, ep, id, data)
    return qry.execute()


# TODO: How to manage auth?
def bulk_create(handle, data, claims=None, ignore_duplicates=False):
    ver, ep = cache.get_version_and_endpoint(handle)
    # check_write_permission(ep, claims)
    # qry = BulkCreate(ver, ep, data, ignore_duplicates=ignore_duplicates)
    # return qry.execute()


# TODO: How to manage auth?
def bulk_update(handle, data, claims=None):
    ver, ep = cache.get_version_and_endpoint(handle)
    # check_write_permission(ep, claims)
    # qry = BulkUpdate(ver, ep, data)
    # return qry.execute()


def list(handle, include=None, filter=None, order=[], limit=None,
         include_limit=None, offset=None, offset_reverse=False,
         url=None):
    ver, ep = cache.get_version_and_endpoint(handle)
    check_access(ep, 'list')

    # Run the query and iterate the rows. PostgreSQL should manage
    # buffering the results, but I'm not 100% sure that will happen.
    # TODO: Confirm result buffering. If it's not happening will need
    #   to allocate a buffer and return chunks?
    # TODO: Ordering.
    qry = List(
        ver,
        ep,
        include=include,
        filter=Filter(filter),
        order=order,
        limit=limit,
        include_limit=include_limit,
        offset=offset,
        offset_reverse=offset_reverse,
        url=url
    )
    for r in qry.iter_rows():
        yield r


def detail(handle, id, include=[]):
    ver, ep = cache.get_version_and_endpoint(handle)
    check_access(ep, 'detail')
    qry = Detail(
        ver,
        ep,
        id,
        include=include
    )
    for r in qry.iter_rows():
        yield r


def delete(handle, id):
    ver, ep = cache.get_version_and_endpoint(handle)
    check_access(ep, 'delete')
    qry = Delete(ver, ep, id)
    qry.execute()


def get(handle, include=[], filter={}, order=[], id=None):
    if id:
        qry = detail(handle, id, include=include)
    else:
        qry = list(handle, include=include, filter=filter, order=order, limit=1)
    for r in qry:
        return r
    raise DoesNotExist



import re

import ujson
from polecat.db.schema.column import ReverseColumn
from polecat.utils import to_bool, to_tuple
from psycopg2.sql import SQL, Composable, Identifier

from .expression import Expression


class Where:
    FILTER_PROG = re.compile(r'^([a-zA-Z][a-zA-Z0-9_]+(?:__[a-zA-Z][a-zA-Z0-9_]+)*)$')
    FILTER_TYPES = None

    def __init__(self, *args, **kwargs):
        self.root = self.parse_input(args, kwargs)

    def get_sql(self, relation):
        self.relation = relation
        if self.root:
            return self.root.get_sql(self)
        else:
            return None

    def parse_input(self, args, kwargs):
        root = None
        for k, v in kwargs.items():
            m = self.FILTER_PROG.match(k)
            if not m:
                raise ValueError(f'Unable to match filter condition: {k}')
            target = m.group(1)
            lookup, flt_cls = self.parse_target(target)
            flt = flt_cls(self, lookup, v)
            if root is None:
                root = flt
            else:
                root = And(root, flt)
        for a in args:
            # TODO: Confirm that `a` is a proper FilterType.
            root = And(root, a)
        return root

    def parse_target(self, target):
        i = target.rfind('__')
        if i != -1:
            try:
                return target[:i], self.FILTER_TYPES[target[i + 2:]]
            except KeyError:
                pass
        return target, Equal

    def merge(self, other, boolean='AND'):
        # TODO: We should really do a check for duplicate filters.
        if self.root:
            if other.root:
                if boolean == 'AND':
                    self.root = And(self.root, other.root)
                else:
                    self.root = Or(self.root, other.root)
        elif other.root:
            self.root = other.root

    def get_primary_columns(self):
        return self.root.get_primary_columns()


class FilterType:
    def __init__(self, filter, lookup, value):
        self.parse_lookup(lookup)
        self.parse_value(filter, value)

    def get_sql(self, filter):
        sql, args = self.eval(filter)
        sql = self.eval_joins(filter, sql)
        return sql, args

    def eval(self, filter):
        pass
        # val = self.value
        # if isinstance(self.value, str):
        #     val = val.format(**filter.context)
        # values.append(val)

    def eval_joins(self, filter, condition):
        if not self.joins:
            return condition
        sql = '%s'
        relation = filter.relation
        args = []
        for i, joined_column_name in enumerate(self.joins):
            # TODO: Handle m2m, reverse fk, reverse m2m.
            column = relation.get_column(joined_column_name)
            if isinstance(column, ReverseColumn):
                prev_tbl_name = relation.alias
                prev_col_name = 'id'
                col_name = column.related_column.name
            else:
                prev_tbl_name = relation.alias
                prev_col_name = column.name
                col_name = 'id'
            relation = column.related_table
            tbl = relation.alias
            # TODO: Use Identifier
            # TODO: PK field other than 'id'.
            next = 'EXISTS (SELECT 1 FROM {} WHERE {} = {} AND %s)'
            args.extend([
                Identifier(tbl),
                SQL('{}.{}').format(Identifier(prev_tbl_name), Identifier(prev_col_name)),
                SQL('{}.{}').format(Identifier(tbl), Identifier(col_name))
            ])
            sql = sql % next
        sql = sql % '{}'
        args.append(condition)
        sql = SQL(sql).format(*args)
        return sql

    def parse_lookup(self, lookup):
        lookup_parts = lookup.split('__')
        if len(lookup_parts) < 1:
            raise ValueError(f'invalid filter: {lookup}')
        # if lookup_parts[-1] in Filter.FILTER_TYPES:
        #     self.type = lookup_parts.pop()
        # else:
        #     self.type = 'eq'
        self.joins = lookup_parts[:-1]
        self.field = lookup_parts[-1]

    def parse_value(self, filter, value):
        # TODO: Oh this isn't nice. I need to be able to use fields to
        # convert values.
        if self.field == 'id':
            try:
                self.value = value.id
            except AttributeError:
                self.value = value
        else:
            self.value = value

    def get_table_column(self, filter):
        relation = filter.relation
        table_name = relation.alias
        for joined_column_name in self.joins:
            column = relation.get_column(joined_column_name)
            relation = column.related_table
            table_name = relation.alias
        return table_name, self.field

    def format(self, format_string, *args):
        # TODO: A little ugly. Now a lot ugly.
        if isinstance(self.value, Composable):
            format_string = format_string % '{}'
            return SQL(format_string).format(*(args + (self.value,))), ()
        elif isinstance(self.value, Expression):
            value_sql, value_args = self.value.to_sql()
            format_string = format_string % '{}'
            return SQL(format_string).format(*(args + (value_sql,))), value_args
        else:
            return SQL(format_string).format(*args), to_tuple(self.get_value(), keep_none=True)

    def get_value(self):
        return self.value

    def get_primary_columns(self):
        # TODO: Test this.
        if self.joins:
            return (self.joins[0],)
        return (self.field,)


class Equal(FilterType):
    def eval(self, filter):
        super().eval(filter)
        try:
            tbl, col = self.get_table_column(filter)
        except KeyError:
            raise ValueError(f'invalid attribute: {self.field}')
        op = '=' if self.value is not None else 'IS'
        return self.format(
            '{}.{} {} %s',
            Identifier(tbl),
            Identifier(col),
            SQL(op)
        )


class NotEqual(FilterType):
    def eval(self, filter):
        super().eval(filter)
        try:
            tbl, col = self.get_table_column(filter)
        except KeyError:
            raise ValueError(f'invalid attribute: {self.field}')
        op = '!=' if self.value is not None else 'IS NOT'
        return self.format('{}.{} {} %s', tbl, col, op)


class Contains(FilterType):
    def eval(self, filter):
        super().eval(filter)
        try:
            tbl, col = self.get_table_column(filter)
        except KeyError:
            raise ValueError(f'invalid attribute: {self.field}')
        return self.format('{}.{} LIKE %s', tbl, col)

    def parse_value(self, filter, value):
        value = '%{}%'.format(value)
        self.value = value.replace('%', r'%%')


class Less(FilterType):
    def eval(self, filter):
        super().eval(filter)
        return self.format('{} < %s', self.field)


class Greater(FilterType):
    def eval(self, filter):
        super().eval(filter)
        return self.format('{} > %s', self.field)


class LessEqual(FilterType):
    def eval(self, filter):
        super().eval(filter)
        return self.format('{} <= %s', self.field)


class GreaterEqual(FilterType):
    def eval(self, filter):
        super().eval(filter)
        return self.format('{} >= %s', self.field)


class In(FilterType):
    def eval(self, filter):
        super().eval(filter)
        try:
            tbl, col = self.get_table_column(filter)
        except KeyError:
            raise ValueError(f'invalid attribute: {self.field}')
        return self.format('{}.{} = ANY (%s)', tbl, col)

    def parse_value(self, filter, value):
        try:
            self.value = ujson.loads(value)
        except Exception:
            raise ValueError(f'Unable to parse "in" filter value: {value}')


class NotIn(In):
    def eval(self, filter):
        FilterType.eval(self, filter)
        return self.format('{} NOT IN %s', self.field)


class IsNull(FilterType):
    def eval(self, filter):
        super().eval(filter)
        try:
            tbl, col = self.get_table_column(filter)
        except KeyError:
            raise ValueError(f'invalid attribute: {self.field}')
        op = 'IS' if self.value else 'IS NOT'
        return self.format(
            '{}.{} {} NULL', Identifier(tbl), Identifier(col), SQL(op)
        )

    def parse_value(self, filter, value):
        self.value = to_bool(value)

    def get_value(self):
        return None


# class NotNull(FilterType):
#     def eval(self, filter):
#         super().eval(filter)
#         return f'{self.field} NOT NULL'


class Overlap(FilterType):
    def eval(self, filter):
        super().eval(filter)
        try:
            tbl, col = self.get_table_column(filter)
        except KeyError:
            raise ValueError(f'invalid attribute: {self.field}')
        return self.format('{}.{} && %s', tbl, col)


class WithinDistance(FilterType):
    def __init__(self, filter, lookup, point, distance):
        super().__init__(filter, lookup, distance)
        self.value = (point, self.value)

    def eval(self, filter):
        super().eval(filter)
        try:
            tbl, col = self.get_table_column(filter)
        except KeyError:
            raise ValueError(f'invalid attribute: {self.field}')
        return self.format('{}.{} <@> %s < %s', tbl, col)


# TODO: This may not be the fastest formulation: https://www.postgresql.org/docs/10/pgtrgm.html#id-1.11.7.41.8
class TrigramSimilar(FilterType):
    def eval(self, filter):
        super().eval(filter)
        try:
            tbl, col = self.get_table_column(filter)
        except KeyError:
            raise ValueError(f'invalid attribute: {self.field}')
        return self.format('{}.{} % %s', tbl, col)


class Operator:
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def get_sql(self, filter):
        raise NotImplementedError

    def eval_sides(self, filter):
        left_sql, left_args = self.left.get_sql(filter)
        right_sql, right_args = self.right.get_sql(filter)
        return left_sql, right_sql, left_args + right_args

    def get_primary_columns(self):
        return self.left.get_primary_columns() + self.right.get_primary_columns()


class And(Operator):
    def get_sql(self, filter):
        left, right, args = self.eval_sides(filter)
        # TODO: Making new SQLs here is probably a tiny bit inefficient.
        if isinstance(self.left, Or):
            left = SQL('({})').format(left)
        if isinstance(self.right, Or):
            right = SQL('({})').format(right)
        return SQL('{} AND {}').format(left, right), args


class Or(Operator):
    def get_sql(self, filter):
        left, right, args = self.eval_sides(filter)
        return SQL('{} OR {}').format(left, right), args


Where.FILTER_TYPES = {
    'eq': Equal,
    'ne': NotEqual,
    'lt': Less,
    'gt': Greater,
    'le': LessEqual,
    'ge': GreaterEqual,
    'in': In,
    'ct': Contains,
    'ni': NotIn,
    'nu': IsNull,
    # 'nn': NotNull,
    'ov': Overlap,
    # 'bt': Between,
    'trigram_similar': TrigramSimilar
}



from factory import base
from polecat.model import default_blueprint
from polecat.model.db import Q

from ...core.context import active_context
from ...model.field import ReverseField
from .field import *  # noqa


class ModelFactory(base.Factory):
    # _options_class = FactoryOptions

    @classmethod
    def _create(cls, model_class, *args, **kwargs):
        inst = model_class(*args, **kwargs)
        Q(inst).insert().into(inst)
        return inst


class FactoryContainer:
    @classmethod
    def get(cls, model):
        return getattr(cls, model.__name__)


def create_model_factory():
    factory = type('Factory', (FactoryContainer,), {})
    for model in default_blueprint.iter_models():
        setattr(
            factory,
            model.__name__,
            create_factory_for_model(model, factory)
        )
    return factory


def create_factory_for_model(model, factory):
    name = f'Auto{model.Meta.name}'
    return type(name, (ModelFactory,), {
        **filter_none({
            f.name: create_factory_field(model, f, factory)
            for f in iter_model_fields(model)
        }),
        'Meta': type('Meta', (), {
            'model': model
        })
    })


def iter_model_fields(model):
    for field in model.Meta.fields.values():
        # TODO: Should test for auto instead.
        if getattr(field, 'primary_key', False):
            continue
        # Don't instantiate reverses by default, gets us into infinite
        # loops.
        if isinstance(field, ReverseField):
            continue
        yield field


@active_context
def create_factory_field(model, field, factory, context=None):
    return (
        context.registries.factory_field_registry[field]
        .get_declaration(model, field, factory)
    )


# TODO: Make this nicer.
def filter_none(d):
    return {
        k: v
        for k, v in d.items()
        if v is not None
    }



import ujson
from polecat.core.config import default_config

from ..event import Event


class APIGatewayRequest:
    def __init__(self, event):
        self.parse_event(event)

    def parse_event(self, event):
        # TODO: Move to config. Also wrong.
        try:
            self.path = event['path']
            self.method = event['httpMethod'].upper()
            self.headers = event.get('headers', {})
            self.host = self.headers.get('Host', None)
            self.query_parameters = event.get('queryStringParameters', None)
            self.body = event.get('body', None)
            self.json = ujson.loads(self.body) if self.body is not None else None
            self.is_valid = True
            # TODO: There must be a better way of doing this.
            if default_config.debug:
                if self.path[:8] == '/LATEST/':
                    self.path = self.path[7:]
                elif self.path[:7] == '/LATEST' and len(self.path) == 7:
                    self.path = '/'
        except KeyError:
            self.is_valid = False
            self.host = None
            self.path = None
            self.method = None
            self.headers = {}


class LambdaEvent(Event):
    def __init__(self, event):
        super().__init__(event)
        self.request = APIGatewayRequest(event)

    def is_http(self):
        return self.request.is_valid

    def is_admin(self):
        return self.event.get('event') == 'admin'

    def get_authorization_header(self):
        return self.request.headers.get('Authorization', None)

    def get_access_control_request_method_header(self):
        return self.request.headers.get('Access-Control-Request-Method', None)



from sanic import Sanic, response
from sanic_cors import CORS

from ...project.project import load_project
from ..event import HttpEvent


class Server:
    def __init__(self, project=None, production=False):
        self.project = project or load_project()
        self.project.prepare()
        self.production = production
        self.port = 80 if production else 8000
        self.sanic_app = Sanic()
        CORS(self.sanic_app, automatic_options=True)
        self.sanic_app.add_route(self.index, '/', methods=['GET'])
        self.sanic_app.add_route(
            self.index,
            '/<path:path>',
            methods=['GET', 'POST', 'PUT', 'PATCH', 'DELETE', 'OPTION']
        )

    async def index(self, request, path=None):
        return self.encode_result(
            await self.project.handle_event(self.request_to_event(request))
        )

    def request_to_event(self, request):
        return HttpEvent(request)

    def encode_result(self, result):
        return response.json(result[0], status=result[1])

    def run(self):
        self.sanic_app.run(host='0.0.0.0', port=self.port)



from polecat.utils.repr import to_repr_attrs

from .utils import Auto
from .variable import SessionVariable


class AsIs:
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return self.value


class ConstructionArguments:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        items = [repr(a) for a in self.args]
        items.append(to_repr_attrs(**self.filter_kwargs()))
        return ', '.join(items)

    def filter_kwargs(self):
        return {
            name: self.map_value(value)
            for name, value in self.kwargs.items()
            if value is not None
        }

    def map_value(self, value):
        if isinstance(value, SessionVariable):
            args = [repr(value.name)]
            if value.type:
                args.append(repr(value.type))
            args = ', '.join(args)
            return AsIs(f'schema.SessionVariable({args})')
        elif value == Auto:
            return AsIs('schema.Auto')
        else:
            return value

    def merge(self, *args, **kwargs):
        self.args += args
        self.kwargs.update({
            k: v
            for k, v in kwargs.items()
            if v is not None
        })
        return self


class Entity:
    def __eq__(self, other):
        return not self.has_changed(other)

    @property
    def dependencies(self):
        return self.get_dependent_entities()

    def get_dependent_entities(self):
        return ()

    def get_construction_arguments(self):
        return ConstructionArguments(args=None, kwargs=None)



import inspect
import logging
from functools import partial

from graphql.type import (GraphQLField, GraphQLInputField,
                          GraphQLInputObjectType, GraphQLInt, GraphQLList,
                          GraphQLNonNull, GraphQLObjectType, GraphQLSchema,
                          GraphQLString)
from polecat.model import default_blueprint

from ..core.context import active_context
from ..model import Model, omit
from ..utils import add_attribute, capitalize, uncapitalize
from ..utils.stringcase import camelcase
# from .field import *  # noqa
from .registry import (add_graphql_create_input, add_graphql_type,
                       add_graphql_update_input, graphql_create_input_registry,
                       graphql_field_registry, graphql_reverse_input_registry,
                       graphql_type_registry, graphql_update_input_registry)
from .resolve import (resolve_all_query, resolve_create_mutation,
                      resolve_delete_mutation, resolve_get_query,
                      resolve_mutation, resolve_query, resolve_update_mutation,
                      resolve_update_or_create_mutation)
from .type import scalars, GraphQLJSON

logger = logging.getLogger(__name__)


def build_graphql_schema():
    # TODO: I need to fix this nonsense. The registries need to be
    # first-class citizens that can be cleared when needed.
    global graphql_create_input_registry
    global graphql_update_input_registry
    global graphql_reverse_input_registry
    all_keys = list(graphql_create_input_registry.keys())
    for k in all_keys:
        del graphql_create_input_registry[k]
    all_keys = list(graphql_update_input_registry.keys())
    for k in all_keys:
        del graphql_update_input_registry[k]
    all_keys = list(graphql_reverse_input_registry.keys())
    for k in all_keys:
        del graphql_reverse_input_registry[k]
    return SchemaBuilder().build()


# TODO: Make the builders class only?
class SchemaBuilder:
    def build(self):
        logger.debug('Building GraphQL schema')
        self.post_build_hooks = []
        self.types = []
        self.queries = {}
        self.mutations = {}
        self.build_models()
        self.build_queries()
        self.build_mutations()
        self.run_post_build_hooks()
        return GraphQLSchema(
            query=GraphQLObjectType(
                name='Query',
                fields=self.queries
            ),
            mutation=GraphQLObjectType(
                name='Mutation',
                fields=self.mutations
            ) if self.mutations else None,
            types=scalars + self.types
        )

    # TODO: Should have called this "delete_input_type".
    @property
    def delete_type(self):
        return self.build_delete_type()

    @property
    def delete_output_type(self):
        return self.build_delete_output_type()

    def build_models(self):
        logger.debug('Building GraphQL models')
        builders = []
        for model in self.iter_models():
            builder_class = getattr(model.Meta, 'builder', ModelBuilder)
            builder = builder_class(self)
            builder.build(model)
            builders.append(builder)
        for builder in builders:
            builder.post_build(self)

    def build_queries(self):
        logger.debug('Building GraphQL queries')
        for query in self.iter_queries():
            builder_class = getattr(query, 'builder', QueryBuilder)
            builder = builder_class(self)
            builder.build(query)

    def build_mutations(self):
        logger.debug('Building GraphQL mutations')
        for mutation in self.iter_mutations():
            builder_class = getattr(mutation, 'builder', MutationBuilder)
            builder = builder_class(self)
            builder.build(mutation)

    def iter_models(self):
        for type in default_blueprint.iter_types():
            if not type.Meta.omit == omit.ALL:
                yield type
        for model in default_blueprint.iter_models():
            if not model.Meta.omit == omit.ALL:
                yield model

    def iter_queries(self):
        for query in default_blueprint.iter_queries():
            yield query

    def iter_mutations(self):
        for mutation in default_blueprint.iter_mutations():
            yield mutation

    def build_delete_type(self):
        # TODO: Use a function cache.
        type = getattr(self, '_delete_type', None)
        if not type:
            type = GraphQLInputObjectType(
                name='DeleteByIDInput',
                fields={
                    'id': GraphQLInputField(GraphQLInt)
                }
            )
            self._delete_type = type
        return type

    def build_delete_output_type(self):
        # TODO: Use a function cache.
        type = getattr(self, '_delete_output_type', None)
        if not type:
            type = GraphQLObjectType(
                name='DeleteByID',
                fields={
                    'id': GraphQLField(GraphQLInt)
                }
            )
            self._delete_output_type = type
        return type

    def run_post_build_hooks(self):
        for hook in self.post_build_hooks:
            hook()


class ModelBuilder:
    def __init__(self, schema_builder):
        self.schema_builder = schema_builder

    def build(self, model):
        self.model = model
        self.type = self.build_type(model)
        self.schema_builder.types.append(self.type)

    # TODO: Can remove `schema_builder`?
    def post_build(self, schema_builder):
        if issubclass(self.model, Model):
            schema_builder.queries.update(self.build_queries(self.model, self.type))
            schema_builder.mutations.update(self.build_mutations(self.model, self.type))

    def build_type(self, model):
        if getattr(model.Meta, 'input', False):
            return InputBuilder(self.schema_builder).build(model)
        else:
            return TypeBuilder(self.schema_builder).build(model)

    def build_queries(self, model, type):
        # TODO: Add the model to the GraphQLField?
        return {
            **self.build_all_queries(model, type),
            **self.build_get_queries(model, type)
        }

    def build_mutations(self, model, type):
        mutations = {}
        if not model.Meta.omit & omit.CREATE:
            args = {}
            create_input = CreateInputBuilder(self.schema_builder).build(model)
            if create_input:
                args['input'] = create_input
            mutations[self.create_mutation_inflection(model)] = GraphQLField(
                type,
                args,
                resolve_create_mutation
            )
        if not model.Meta.omit & omit.UPDATE:
            args = {
                'id': GraphQLNonNull(GraphQLInt)
            }
            update_input = UpdateInputBuilder(self.schema_builder).build(model)
            if update_input:
                args['input'] = update_input
            mutations[self.update_mutation_inflection(model)] = GraphQLField(
                type,
                args,
                resolve_update_mutation
            )
        if not model.Meta.omit & omit.UPDATE and not model.Meta.omit & omit.CREATE:
            args = {
                'id': GraphQLInt
            }
            if update_input:
                args['input'] = update_input
            mutations[self.update_or_create_mutation_inflection(model)] = GraphQLField(
                type,
                args,
                resolve_update_or_create_mutation
            )
        if not model.Meta.omit & omit.DELETE:
            # TODO: Need to use type instead of delete_output_type
            # because it contains details about the model class
            # used. Will need to think about how to handle situations
            # where a type is shared amongst numerous models.
            mutations[self.delete_mutation_inflection(model)] = GraphQLField(
                type,
                {
                    'input': DeleteInputBuilder(self.schema_builder).build(model)
                },
                resolve_delete_mutation
            )
        return mutations

    def build_all_queries(self, model, type):
        queries = {}
        if not model.Meta.omit & omit.LIST:
            queries[self.all_query_inflection(model)] = GraphQLField(
                GraphQLList(type),
                resolve=resolve_all_query
            )
        return queries

    def build_get_queries(self, model, type):
        queries = {}
        if not model.Meta.omit & omit.GET:
            # TODO: Need to add some details to the field to support
            # getting by?
            for name, field in model.Meta.cc_fields.items():
                if field.omit & omit.GET:
                    continue
                if not field.use_get_query():
                    continue
                queries[self.get_query_inflection(model, field)] = add_attribute(
                    GraphQLField(
                        type,
                        {
                            # TODO: What happens when the type linked is
                            # something complicated, like a GraphQLObject?
                            # Will need to use the PK of the
                            # relationship. What if there's several?
                            name: type.fields[name].type
                        },
                        resolve=resolve_get_query
                    ),
                    '_type', type
                )
        return queries

    def all_query_inflection(self, model):
        # TODO: Not sure I need to capitalize?
        return f'all{model.Meta.plural}'

    def get_query_inflection(self, model, field):
        name = f'get{model.Meta.name}'
        if not field.primary_key:
            name += f'By{capitalize(field.cc_name)}'
        return name

    def create_mutation_inflection(self, model):
        return f'create{model.Meta.name}'

    def update_mutation_inflection(self, model):
        return f'update{model.Meta.name}'

    def update_or_create_mutation_inflection(self, model):
        return f'updateOrCreate{model.Meta.name}'

    def delete_mutation_inflection(self, model):
        return f'delete{model.Meta.name}'


class TypeBuilder:
    object_type_class = GraphQLObjectType

    def __init__(self, schema_builder):
        self.schema_builder = schema_builder

    def build(self, model):
        object_type_class = self.get_object_type_class()
        # TODO: Description?
        type = add_attribute(
            object_type_class(
                name=self.get_type_name(model),
                fields=partial(self.build_all_fields, model)
            ),
            '_model', model
        )
        # TODO: Current assumption is that there's just one gql type
        # per model. May need to alter this in the future. I'm not
        # storing any of the model create, update, etc inputs.
        self.register_type(model, type)
        return type

    def get_type_name(self, model):
        return model.Meta.name

    def build_all_fields(self, model):
        logger.debug(f'Building GraphQL fields for model {model.Meta.name} using builder {self.__class__.__name__}')
        # TODO: Can I make this more functional?
        fields = {}
        for name, field in self.iter_fields(model):
            result = self.build_field(model, field)
            if isinstance(result, dict):
                fields.update(result)
            elif result is None:
                pass
            else:
                fields[name] = result
            self.add_hooks_from_field(model, field, name)
        return fields

    def add_hooks_from_field(self, model, field, field_name):
        hooks = field.graphql_post_build_hooks
        self.schema_builder.post_build_hooks.extend([
            partial(h, self.schema_builder, model, field, field_name)
            for h in hooks
        ])

    def build_field(self, model, field):
        for base_class in inspect.getmro(field.__class__):
            my_graphql_field = graphql_field_registry.get(base_class)
            # TODO: Nomenclature mishap here. Need a name for my
            # GraphQL type abstraction that doesn't clash with actual
            # graphql types.
            if my_graphql_field:
                logger.debug(f'Building field {field.cc_name}')
                field = self.get_graphql_field(model, field, my_graphql_field)
                logger.debug('DONE')
                return field
        raise Exception(f'unknown field type {field}')

    def get_graphql_field(self, model, field, my_graphql_field):
        # TODO: Maybe remove the function call and do it in __init__?
        return my_graphql_field(model, field).make_graphql_field(self.schema_builder)

    def get_object_type_class(self):
        return self.object_type_class

    def register_type(self, model, type):
        add_graphql_type(model, type)

    def iter_fields(self, model):
        for name, field in model.Meta.cc_fields.items():
            if field.omit & (omit.LIST | omit.GET):
                continue
            yield name, field


class InputBuilder(TypeBuilder):
    object_type_class = GraphQLInputObjectType

    def get_type_name(self, model):
        # TODO: Probably shouldn't need this conditional.
        name = model.Meta.name
        if name[-5:] != 'Input':
            name += 'Input'
        return name

    def get_graphql_field(self, model, field, my_graphql_field):
        # TODO: Maybe remove the function call and do it in __init__?
        return my_graphql_field(model, field, input=True).make_graphql_field(self.schema_builder)

    def iter_fields(self, model):
        for name, field in model.Meta.cc_fields.items():
            yield name, field


class CreateInputBuilder(InputBuilder):
    def get_type_name(self, model):
        return f'{model.Meta.name}CreateInput'

    def get_graphql_field(self, model, field, my_graphql_field):
        # TODO: Maybe remove the function call and do it in __init__?
        # TODO: Have to pass in a type because reverse fields generate
        # a new object-type internally instead of just a type, so it
        # needs a name. So, if I don't pass in a type string create,
        # update, delete will produce identically named object-types.
        return my_graphql_field(
            model,
            field,
            input=True,
            registry=graphql_create_input_registry  # TODO: Yuck
        ).make_graphql_field(self.schema_builder)

    def register_type(self, model, type):
        add_graphql_create_input(model, type)

    def iter_fields(self, model):
        for name, field in model.Meta.cc_fields.items():
            if field.omit & omit.CREATE:
                continue
            yield name, field


class UpdateInputBuilder(InputBuilder):
    def get_type_name(self, model):
        return f'{model.Meta.name}UpdateInput'

    def get_graphql_field(self, model, field, my_graphql_field):
        # TODO: Maybe remove the function call and do it in __init__?
        # TODO: Check if it's not the ID field used for the update,
        # and if not, make the field optional.
        # TODO: Have to pass in a type because reverse fields generate
        # a new object-type internally instead of just a type, so it
        # needs a name. So, if I don't pass in a type string create,
        # update, delete will produce identically named object-types.
        return my_graphql_field(
            model,
            field,
            input=True,
            registry=graphql_update_input_registry  # TODO: Yuck
        ).make_graphql_field(self.schema_builder)

    def register_type(self, model, type):
        add_graphql_update_input(model, type)

    def iter_fields(self, model):
        for name, field in model.Meta.cc_fields.items():
            if field.omit & omit.UPDATE:
                continue
            if name == 'id':
                continue
            yield name, field


class DeleteInputBuilder(InputBuilder):
    def build(self, model):
        object_type_class = self.get_object_type_class()
        # TODO: Description?
        type = add_attribute(
            object_type_class(
                name=self.get_type_name(model),
                fields={
                    'id': GraphQLInt
                }
            ),
            '_model', model
        )
        # TODO: Current assumption is that there's just one gql type
        # per model. May need to alter this in the future. I'm not
        # storing any of the model create, update, etc inputs.
        # self.register_type(model, type)
        return type

    def get_type_name(self, model):
        return f'{model.Meta.name}DeleteInput'

    def iter_fields(self, model):
        for name, field in model.Meta.cc_fields.items():
            if field.omit & omit.DELETE:
                continue
            yield name, field


class ReverseModelInputBuilder(InputBuilder):
    def build(self, model, source_field):
        # TODO: Do I need to pass source_field around, or okay to
        # store on the builder?
        self.source_field = source_field
        logger.debug(f'Building GraphQL reverse model input {self.get_type_name(model)}')
        return super().build(model)

    def get_graphql_field(self, model, field, my_graphql_field):
        # TODO: Maybe remove the function call and do it in __init__?
        # TODO: Check if it's not the ID field used for the update,
        # and if not, make the field optional.
        # TODO: Have to pass in a type because reverse fields generate
        # a new object-type internally instead of just a type, so it
        # needs a name. So, if I don't pass in a type string create,
        # update, delete will produce identically named object-types.
        return my_graphql_field(
            model,
            field,
            input=True,
            registry=graphql_create_input_registry  # TODO: Yuck
        ).make_graphql_field(self.schema_builder)

    def get_type_name(self, model):
        return f'{model.Meta.name}{capitalize(self.source_field.cc_name)}ReverseInput'

    def iter_fields(self, model):
        for name, field in model.Meta.cc_fields.items():
            if field != self.source_field:
                yield name, field


class QueryBuilder:
    def __init__(self, schema_builder):
        self.schema_builder = schema_builder

    def build(self, query):
        self.schema_builder.queries.update(self.build_queries(query))

    def build_queries(self, query):
        queries = {}
        if not query.Meta.omit & omit.ALL:
            queries[self.query_inflection(query)] = add_attribute(
                GraphQLField(
                    self.build_return_type(query),
                    self.build_arguments(query),
                    resolve_query
                ),
                '_query', query
            )
        return queries

    def build_return_type(self, query):
        if isinstance(query.returns, list):
            return GraphQLList(graphql_type_registry[query.returns[0]])
        else:
            return graphql_type_registry[query.returns]

    def build_arguments(self, query):
        args = {}
        for name, type in getattr(query, 'arguments', ()):
            # TODO: This should be looking up from a table.
            args[camelcase(name)] = self.map_type(type)
        return args

    def map_type(self, type):
        if type == str:
            return GraphQLString
        elif type == int:
            return GraphQLInt
        elif type == dict:
            return GraphQLJSON
        else:
            raise NotImplementedError(f'invalid GraphQL argument: {type}')

    def query_inflection(self, query):
        return uncapitalize(query.Meta.name)


class MutationBuilder:
    def __init__(self, schema_builder):
        self.schema_builder = schema_builder

    def build(self, mutation):
        self.schema_builder.mutations.update(self.build_mutations(mutation))

    def build_mutations(self, mutation):
        mutations = {}
        if not mutation.omit & omit.ALL:
            mutations[self.mutation_inflection(mutation)] = add_attribute(
                GraphQLField(
                    graphql_type_registry[mutation.returns],
                    {
                        'input': graphql_type_registry[mutation.input]
                    } if mutation.input else None,
                    resolve_mutation
                ),
                '_mutation', mutation
            )
        return mutations

    def mutation_inflection(self, mutation):
        return uncapitalize(mutation.name)



import importlib
import os
import subprocess
import sys
from pathlib import Path
from shutil import copytree, make_archive
from tempfile import TemporaryDirectory

from ...utils.exceptions import KnownError

server_main_py = '''# Auto-generated by Polecat

from polecat.deploy.aws.server import LambdaServer

server = LambdaServer()


def handler(event, context):
    return server.handle(event, context)
'''


def build(project, source=None, local_packages=None, feedback=None):
    feedback.declare_steps(
        # 'Update docker images',
        'Install polecat',
        'Install dependencies',
        'Create archive'
    )
    # with feedback('Update docker images'):
    #     subprocess.check_output('docker pull python:3.7'.split(), stderr=subprocess.PIPE)
    source = Path(source) if source else (Path.cwd() / project)
    local_packages = local_packages or []
    with TemporaryDirectory() as root:
        root = Path(root)
        if not source.exists():
            # TODO: Don't like this known error stuff.
            raise KnownError('source path does not exist')
        with feedback('Install polecat') as fb:
            # TODO: Find a way to automatically set this version.
            # install_package('polecat==0.0.5', Path(root))
            for pkg in find_polecat_package(fb):
                install_package(pkg, Path(root))
        with feedback('Install dependencies') as fb:
            for lpkg in local_packages:
                for pkg in find_package(lpkg, fb):
                    install_package(pkg, Path(root))
        with feedback('Create archive') as fb:
            copytree(source, root / os.path.basename(source))
            with open(root / 'main.py', 'wt') as f:
                f.write(server_main_py)
            make_archive('server', 'zip', str(root))


def install_package(package, root):
    if str(package).startswith('http'):
        cmd = f'python -m pip install --target /app {package}'
        docker_cmd = f'docker run --rm --user 1000:1000 -v {root}:/app python:3.7'
    else:
        cmd = 'python -m pip install --target /app /tmp/pkg'
        docker_cmd = f'docker run --rm --user 1000:1000 -v {package}:/tmp/pkg -v {root}:/app python:3.7'
    split_cmd = f'{docker_cmd} {cmd}'.split()
    subprocess.check_output(split_cmd, stderr=subprocess.PIPE)


def find_polecat_package(feedback):
    try:
        import polecat
    except ImportError:
        return ('polecat',)
    root = Path(polecat.__file__).parent.parent
    setup_file = root / 'setup.py'
    if setup_file.is_file():
        feedback.add_notice('Using local package polecat')
        # TODO: So long as the current pip version of graphql-server does not
        # use graphql-core-next we need to manually install it.
        return (root, 'https://github.com/norman-thomas/graphql-server-core/tarball/master')
    else:
        return ('polecat',)


def find_package(package, feedback):
    try:
        module = importlib.import_module(package)
    except ImportError:
        return (package,)
    root = Path(module.__file__).parent.parent
    setup_file = root / 'setup.py'
    if setup_file.is_file():
        feedback.add_notice(f'Using local package {package}')
        return (root,)
    else:
        return (package,)



from polecat.utils import substitute
from polecat.utils.feedback import feedback
from termcolor import colored

from .constants import PROJECT_PREFIX, PROJECT_REGISTRY
from .exceptions import EntityDoesNotExist, EntityExists, KnownError
from .operations import (attach_group_policy, create_group, create_policy,
                         delete_group, delete_parameter, delete_policy,
                         detach_group_policy, find_policy, get_parameter,
                         set_parameter)
from .policy import project_policy
from .utils import aws_client


@feedback
def create_project(project, bucket, feedback):
    iam = aws_client('iam')
    ssm = aws_client('ssm')
    with feedback(f'Create project {colored(project, "blue")}'):
        add_project_to_registry(project, ssm=ssm)
    with feedback(f'Create project group'):
        create_project_group(project, iam=iam)
    with feedback(f'Create project policy'):
        create_project_group_policy(project, bucket, iam=iam)
    with feedback(f'Attach project group'):
        attach_project_group(project, iam=iam)


@feedback
def list_projects(feedback):
    ssm = aws_client('ssm')
    with feedback(f'List projects'):
        projects = get_parameter(PROJECT_REGISTRY, default=[], ssm=ssm)
        return [p for p in projects if p != '_']


@feedback
def delete_project(project, feedback):
    iam = aws_client('iam')
    ssm = aws_client('ssm')
    assert_project_exists(project, ssm=ssm)
    with feedback(f'Delete project {colored(project, "blue")}'):
        assert_project_is_empty(project, ssm=ssm)
        # TODO: Need to fetch the path, then delete all of them.
        delete_parameter(PROJECT_PREFIX.format(project), ssm=ssm)
        delete_project_from_registry(project, ssm=ssm)
    with feedback(f'Detach project policy'):
        detach_project_group(project, iam=iam)
    with feedback(f'Delete project policy'):
        delete_project_group_policy(project, iam=iam)
    with feedback(f'Delete project group'):
        delete_project_group(project, iam=iam)


def add_project_to_registry(project, ssm=None):
    registry = get_parameter(PROJECT_REGISTRY, ssm=ssm)
    if project in registry:
        raise EntityExists
    try:
        registry.remove('_')
    except ValueError:
        pass
    registry.append(project)
    set_parameter(PROJECT_REGISTRY, registry)


def delete_project_from_registry(project, ssm=None):
    registry = get_parameter(PROJECT_REGISTRY, ssm=ssm)
    registry.remove(project)
    if not registry:
        registry.append('_')
    set_parameter(PROJECT_REGISTRY, registry)


def create_project_group(project, iam=None):
    create_group(project, iam=iam)


def delete_project_group(project, iam=None):
    delete_group(project, iam=iam)


def create_project_group_policy(project, bucket, iam=None):
    create_policy(
        project,
        substitute(project_policy, project=project, bucket=bucket),
        iam=iam
    )


def delete_project_group_policy(project, iam=None):
    delete_policy(project, iam=iam)


def attach_project_group(project, iam=None):
    attach_group_policy(project, find_policy(project).arn, iam=iam)


def detach_project_group(project, iam=None):
    detach_group_policy(project, find_policy(project).arn, iam=iam)


def project_exists(project, ssm=None):
    registry = get_parameter(PROJECT_REGISTRY, ssm=ssm)
    return project in registry


def assert_project_exists(project, ssm=None):
    if not project_exists(project, ssm=ssm):
        raise EntityDoesNotExist(
            f'Project {colored(project, "blue")} does not exist'
        )


def assert_project_is_empty(project, ssm=None):
    param = PROJECT_PREFIX.format(project) + 'deploymentRegistry'
    deployments = get_parameter(param, ssm=ssm)
    if len(deployments) > 1 or (len(deployments) == 1 and deployments[0] != '_'):
        raise KnownError('project is not empty')
